"""Tests for `pixell` package."""

import unittest

from pixell import enmap
from pixell import curvedsky
from pixell import lensing
from pixell import array_ops
from pixell import enplot
from pixell import powspec
from pixell import reproject
from pixell import pointsrcs
from pixell import wcsutils
from pixell import utils as u
from pixell import colors
from pixell import fft
from pixell import tilemap
from pixell import utils
import numpy as np
import pickle
import os,sys,time

import matplotlib
matplotlib.use('Agg')
import numpy as np
import itertools,yaml,pickle,os,sys
import matplotlib.pyplot as plt

TEST_DIR = os.path.dirname(__file__)
DATA_PREFIX = os.path.join(TEST_DIR, 'data/')

def get_reference_pixels(shape):
    """For a given 2D array, return a list of pixel indices
    corresponding to locations of a pre-determined and fixed
    pattern of reference pixels.

    e.g even x even
    1100110011
    1100110011
    0000000000
    0000000000
    1100110011
    1100110011
    0000000000
    0000000000
    1100110011
    1100110011

    e,g. odd x odd
    110010011
    110010011
    000000000
    000000000
    110010011
    000000000
    000000000
    110010011
    110010011

    e.g even x odd
    110010011
    110010011
    000000000
    000000000
    110010011
    110010011
    000000000
    000000000
    110010011
    110010011

    requires N>=5 in each axis
    """
    Ny,Nx = shape[-2:]
    assert (Ny>=5) and (Nx>=5), "Tests are not implemented for arrays with a dimension<5."
    """Given N, return 0,1,{x},N-2,N-1, where {x} is N//2-1,N//2 if N is even
    and {x} is N//2 if N is odd.
    """
    midextremes = lambda N: [0,1,N//2-1,N//2,N-2,N-1] if N%2==0 else [0,1,N//2,N-2,N-1]
    ys = midextremes(Ny)
    xs = midextremes(Nx)
    pixels = np.array(list(itertools.product(ys,xs)))
    return pixels

def mask(arr,pixels,val=0):
    """Mask an array arr based on array pixels of (y,x) pixel coordinates of (Npix,2)"""
    arr[...,pixels[:,0],pixels[:,1]] = val
    return arr

def get_pixel_values(arr,pixels):
    """Get values of arr at pixels specified in pixels (Npix,2)"""
    return arr[...,pixels[:,0],pixels[:,1]]

def get_meansquare(arr):
    return np.mean(arr*2.)

def save_mask_image(filename,shape):
    """Save a minimal plot of an array masked by the currently implemented reference
    pixel geometry

    e.g.
    > shape = (11,12)
    > save_mask_image("test_mask.png",shape)
    """
    arr = np.zeros(shape)
    pixels = get_reference_pixels(shape)
    masked = mask(arr,pixels,val=1)
    fig = plt.figure()
    im = plt.imshow(masked,cmap='rainbow')
    ax = plt.gca()
    ax.set_xticks(np.arange(0,shape[1])+0.5);
    ax.set_yticks(np.arange(0,shape[0])+0.5);
    ax.grid(which='major',color='w', linestyle='-', linewidth=5)
    ax.tick_params(axis='x', colors=(0,0,0,0))
    ax.tick_params(axis='y', colors=(0,0,0,0))
    for spine in im.axes.spines.values():
        spine.set_edgecolor((0,0,0,0))
    plt.savefig(filename, bbox_inches='tight')

def get_spectrum(ntype,noise,lmax,lmax_pad):
    ells = np.arange(0,lmax+lmax_pad)
    if ntype=="white": return np.ones(shape=(ells.size,))*(noise**2.)*((np.pi/180./60.)**2.)
    if ntype=="white_dl":
        spec = np.zeros(shape=(ells.size,))
        spec[2:] = (noise**2.)*((np.pi/180./60.)**2.)*2.*np.pi/ells[2:]/(ells+1.)[2:]
        return spec
    raise NotImplementedError

def get_spectra(yml_section,lmax,lmax_pad):
    spectra = {}
    for s in yml_section:
        spectra[s['name']] = get_spectrum(s['type'],s['noise'],lmax,lmax_pad)
    return spectra

def get_geometries(yml_section):
    geos = {}
    for g in yml_section:
        if g['type']=='fullsky':
            geos[g['name']] = enmap.fullsky_geometry(res=np.deg2rad(g['res_arcmin']/60.),proj=g['proj'],variant="CC")
        elif g['type']=='pickle':
            geos[g['name']] = pickle.load(open(DATA_PREFIX+"%s"%g['filename'],'rb'))
        else:
            raise NotImplementedError
    return geos

def generate_map(shape,wcs,powspec,lmax,seed):
    return curvedsky.rand_map(shape, wcs, powspec, lmax=lmax, dtype=np.float64, seed=seed, spin=[0,2], method="auto", verbose=False)

def check_equality(imap1,imap2):
    assert np.all(imap1.shape==imap2.shape)
    assert wcsutils.equal(imap1.wcs,imap2.wcs)
    try:
        assert np.all(np.isclose(imap1,imap2))
    except:
        from orphics import io
        io.plot_img(imap1,"i1.png",lim=[-1.5,2])
        io.plot_img(imap2,"i2.png",lim=[-1.5,2])
        io.plot_img((imap1-imap2)/imap1,"ip.png",lim=[-0.1,0.1])
        assert 1==0

    
def get_extraction_test_results(yaml_file):
    print("Starting tests from ",yaml_file)
    with open(yaml_file) as f:
        config = yaml.safe_load(f)
    geos = get_geometries(config['geometries'])
    lmax = config['lmax'] ; lmax_pad = config['lmax_pad']
    spectra = get_spectra(config['spectra'],lmax,lmax_pad)
    seed = config['seed']

    results = {}
    for g in geos.keys():
        results[g] = {}
        for s in spectra.keys():
            results[g][s] = {}
            imap = generate_map(geos[g][0][-2:],geos[g][1],spectra[s],lmax,seed)

            # Do write and read test
            filename = "temporary_map.fits" # NOT THREAD SAFE
            enmap.write_map(filename,imap)
            imap_in = enmap.read_map(filename)
            check_equality(imap,imap_in)
            for e in config['extracts']:
                print("Doing test for extract ",e['name']," with geometry ",g," and spectrum ",s,"...")
                if e['type']=='slice':
                    box = np.deg2rad(np.array(e['box_deg']))
                    cutout = enmap.read_map(filename,box=box)
                    cutout_internal = imap.submap(box=box)
                check_equality(cutout,cutout_internal)
                pixels = get_reference_pixels(cutout.shape)
                results[g][s]['refpixels'] = get_pixel_values(cutout,pixels)
                results[g][s]['meansquare'] = get_meansquare(cutout)

    os.remove(filename)
    return results,config['result_name']

lens_version = '071123'

def get_offset_result(res=1.,dtype=np.float64,seed=1):
    shape,wcs  = enmap.fullsky_geometry(res=np.deg2rad(res), variant="CC")
    shape = (3,) + shape
    obs_pos = enmap.posmap(shape, wcs)
    np.random.seed(seed)
    grad = enmap.enmap(np.random.random(shape),wcs)*1e-3
    raw_pos = enmap.samewcs(lensing.offset_by_grad(obs_pos, grad, pol=shape[-3]>1, geodesic=True), obs_pos)
    return obs_pos,grad,raw_pos

def get_lens_result(res=1.,lmax=400,dtype=np.float64,seed=1):
    shape,wcs  = enmap.fullsky_geometry(res=np.deg2rad(res), variant="CC")
    shape = (3,) + shape
    # ells = np.arange(lmax)
    ps_cmb,ps_lens = powspec.read_camb_scalar(DATA_PREFIX+"test_scalCls.dat")
    ps_lensinput = np.zeros((4,4,ps_cmb.shape[-1]))
    ps_lensinput[0,0] = ps_lens
    ps_lensinput[1:,1:] = ps_cmb
    lensed = lensing.rand_map(shape, wcs, ps_lensinput, lmax=lmax, maplmax=None, dtype=dtype, seed=seed, phi_seed=None, spin=[0,2], output="lu", geodesic=True, verbose=False, delta_theta=None)
    return lensed

# Helper functions for adjointness tests
def zip_alm(alm, ainfo):
    n = ainfo.lm2ind(1,1)
    first  = alm[...,:n].real
    second = alm[...,n:].view(utils.real_dtype(alm.dtype))*2**0.5
    return np.concatenate([first, second],-1)

def unzip_alm(zalm, ainfo):
    n = ainfo.lm2ind(1,1)
    oalm = np.zeros(zalm.shape[:-1] + (ainfo.nelem,), utils.complex_dtype(zalm.dtype))
    oalm[...,:n] = zalm[...,:n]
    oalm[...,n:] = zalm[...,n:].view(oalm.dtype)/2**0.5
    return oalm

def zalm_len(ainfo): return (2*ainfo.nelem-ainfo.lm2ind(1,1)).astype(int)
def zip_mat(mat):
    # Mat is ncomp_alm,nzalm,ncomp_map,ny,nx.
    # Want (ncomp*ncomp*nzalm,ny,nx)
    mat = np.moveaxis(mat, 2, 1)
    mat = mat.reshape((-2,)+mat.shape[-2:])
    return mat

def map_bash(fun, shape, wcs, ncomp, lmax, dtype=np.float64):
    ctype = utils.complex_dtype(dtype)
    ainfo = curvedsky.alm_info(lmax)
    nzalm = zalm_len(ainfo)
    umap  = enmap.zeros((ncomp,)+shape, wcs, dtype=dtype)
    oalm  = np.zeros((ncomp,ainfo.nelem), dtype=ctype)
    mat   = np.zeros((ncomp,nzalm,ncomp)+shape, dtype=dtype)
    for I in utils.nditer((ncomp,)+shape):
        umap[I] = 1
        oalm[:] = 0
        fun(map=umap, alm=oalm, ainfo=ainfo)
        mat[(slice(None),slice(None))+I] = zip_alm(oalm, ainfo)
        umap[I] = 0
    return zip_mat(mat)

def alm_bash(fun, shape, wcs, ncomp, lmax, dtype=np.float64):
    ctype = utils.complex_dtype(dtype)
    ainfo = curvedsky.alm_info(lmax)
    nzalm = zalm_len(ainfo)
    zalm  = np.zeros((ncomp,nzalm), dtype)
    omap  = enmap.zeros((ncomp,)+shape, wcs, dtype)
    mat   = np.zeros((ncomp,nzalm,ncomp)+shape, dtype)
    for ci in range(ncomp):
        for i in range(nzalm):
            # Why is this 0.5 needed?
            zalm[ci,i] = 1 #if i < ainfo.lm2ind(1,1) else 0.5
            omap[:] = 0
            fun(alm=unzip_alm(zalm,ainfo), map=omap, ainfo=ainfo)
            mat[ci,i] = omap
            zalm[ci,i] = 0
    return zip_mat(mat)

# End of adjointness helpers

class PixelTests(unittest.TestCase):


    def test_almxfl(self):
        import healpy as hp

        for lmax in [100,400,500,1000]:
            ainfo = curvedsky.alm_info(lmax)
            alms = hp.synalm(np.ones(lmax+1),lmax = lmax, new=True)
            filtering = np.ones(lmax+1)
            alms0 = ainfo.lmul(alms.copy(),filtering)
            assert np.all(np.isclose(alms0,alms))

        for lmax in [100,400,500,1000]:
            ainfo = curvedsky.alm_info(lmax)
            alms = hp.synalm(np.ones(lmax+1),lmax = lmax, new=True)
            alms0 = curvedsky.almxfl(alms.copy(),lambda x: np.ones(x.shape))
            assert np.all(np.isclose(alms0,alms))
            
        

    def test_rand_alm(self):
        def nalm(lmax):
            return (lmax + 1) * (lmax + 2) / 2

        lmaxes = [50, 100, 150, 300]
        
        mypower = np.ones(50)
        for lmax in lmaxes:
            palm = curvedsky.rand_alm(mypower, lmax = lmax)
            halm = curvedsky.rand_alm_healpy(  mypower, lmax = lmax)
            
            print("nalm(%i) = %i, curvedsky.rand_alm gives %s, curvedsky.rand_alm_healpy gives %s "\
	              % (lmax, \
                     nalm(lmax),\
                     palm.shape, \
                     halm.shape)        )
            assert np.all(np.isclose(np.asarray(palm.shape),np.asarray(halm.shape)))
            
    
    def test_offset(self):
        obs_pos,grad,raw_pos = get_offset_result(1.)
        obs_pos0 = enmap.read_map(DATA_PREFIX+"MM_offset_obs_pos_%s.fits" % lens_version)
        grad0 = enmap.read_map(DATA_PREFIX+"MM_offset_grad_%s.fits"  % lens_version)
        raw_pos0 = enmap.read_map(DATA_PREFIX+"MM_offset_raw_pos_%s.fits"  % lens_version)
        assert np.all(np.isclose(obs_pos,obs_pos0))
        assert np.all(np.isclose(raw_pos,raw_pos0))
        assert np.all(np.isclose(grad,grad0))
        assert wcsutils.equal(grad.wcs,grad0.wcs)
        assert wcsutils.equal(obs_pos.wcs,obs_pos0.wcs)
        assert wcsutils.equal(raw_pos.wcs,raw_pos0.wcs)

    def test_lensing(self):
        lensed,unlensed = get_lens_result(1.,400,np.float64)
        lensed0 = enmap.read_map(DATA_PREFIX+"MM_lensed_%s.fits"  % lens_version)
        unlensed0 = enmap.read_map(DATA_PREFIX+"MM_unlensed_%s.fits"  % lens_version)
        y,x = lensed0.posmap()
        assert np.all(np.isclose(lensed,lensed0))
        assert np.all(np.isclose(unlensed,unlensed0))
        assert wcsutils.equal(lensed.wcs,lensed0.wcs)
        assert wcsutils.equal(unlensed.wcs,unlensed0.wcs)
        assert wcsutils.equal(unlensed.wcs,lensed.wcs)
    
    def test_enplot(self):
        print("Testing enplot...")
        shape,wcs = enmap.geometry(pos=(0,0),shape=(3,100,100),res=0.01)
        a = enmap.ones(shape,wcs)
        # basic
        p = enplot.plot(a)
        # colorbar
        p = enplot.plot(a, colorbar=True)
        # annotation
        p = enplot.plot(a, annotate=DATA_PREFIX+"annot.txt")

    def test_fft(self):
        # Tests that ifft(ifft(imap))==imap, i.e. default normalizations are consistent
        shape,wcs = enmap.geometry(pos=(0,0),shape=(3,100,100),res=0.01)
        imap = enmap.enmap(np.random.random(shape),wcs)
        assert np.all(np.isclose(imap,enmap.ifft(enmap.fft(imap,normalize='phy'),normalize='phy').real))
        assert np.all(np.isclose(imap,enmap.ifft(enmap.fft(imap)).real))

    def test_fft_input_shape(self):
        # Tests fft for various shapes and choices of axes.
        # 1D FFT over last axis for 3d array.
        signal = np.ones((1, 2, 5))
        signal[0,1,:] = 10.
        out_exp = np.zeros((1, 2, 5), dtype=np.complex128)
        out_exp[0,0,0] = 5
        out_exp[0,1,0] = 50
        out = fft.fft(signal)
        np.testing.assert_allclose(out, out_exp, atol=1e-12)
        self.assertTrue(out.flags['C_CONTIGUOUS'])

        # 1D FFT over middle axis for 3d array.
        signal = np.ones((1, 5, 2))
        signal[0,:,1] = 10.
        out_exp = np.zeros((1, 5, 2), dtype=np.complex128)
        out_exp[0,0,0] = 5
        out_exp[0,0,1] = 50
        out = fft.fft(signal, axes=[-2])
        np.testing.assert_allclose(out, out_exp, atol=1e-12)
        self.assertTrue(out.flags['C_CONTIGUOUS'])

        # 2D FFT over last 2 axes of 4d array.
        signal = np.ones((1, 2, 5, 10))
        signal[0,1,:] = 10.
        out_exp = np.zeros((1, 2, 5, 10), dtype=np.complex128)
        out_exp[0,0,0,0] = 50
        out_exp[0,1,0,0] = 500
        out = fft.fft(signal, axes=[-2, -1])
        np.testing.assert_allclose(out, out_exp, atol=1e-12)
        self.assertTrue(out.flags['C_CONTIGUOUS'])

        # 2D FFT over last 2 axes of 4d non-contiguous array.
        signal = np.ones((1, 2, 5, 10), dtype=np.complex128)
        signal[0,1,:] = 10
        ft = np.zeros((5, 10, 1, 2), dtype=np.complex128).transpose(2, 3, 0, 1)
        out_exp = np.zeros_like(ft)
        out_exp[0,0,0,0] = 50
        out_exp[0,1,0,0] = 500
        out = fft.fft(signal, ft=ft, axes=[-2, -1])
        np.testing.assert_allclose(out, out_exp, atol=1e-12)
        self.assertTrue(np.shares_memory(ft, out))
        self.assertFalse(out.flags['C_CONTIGUOUS'])

        # 2D FFT over middle 2 axes of 4d array.
        signal = np.ones((1, 5, 10, 2))
        signal[0,:,:,1] = 10.
        out_exp = np.zeros((1, 5, 10, 2), dtype=np.complex128)
        out_exp[0,0,0,0] = 50
        out_exp[0,0,0,1] = 500
        out = fft.fft(signal, axes=[-3, -2])
        np.testing.assert_allclose(out, out_exp, atol=1e-12)
        self.assertTrue(out.flags['C_CONTIGUOUS'])

    def test_ifft_input_shape(self):
        # Tests ifft for various shapes and choices of axes.
        # 1D IFFT over last axis for 3d array.
        fsignal = np.ones((1, 2, 5), dtype=np.complex128)
        fsignal[0,1,:] = 10.
        out_exp = np.zeros((1, 2, 5))
        out_exp[0,0,0] = 5
        out_exp[0,1,0] = 50
        out = fft.ifft(fsignal)
        np.testing.assert_allclose(out, out_exp, atol=1e-12)
        self.assertTrue(out.flags['C_CONTIGUOUS'])

        # 1D IFFT over middle axis for 3d array.
        fsignal = np.ones((1, 5, 2), dtype=np.complex128)
        fsignal[0,:,1] = 10.
        out_exp = np.zeros((1, 5, 2))
        out_exp[0,0,0] = 5
        out_exp[0,0,1] = 50
        out = fft.ifft(fsignal, axes=[-2])
        np.testing.assert_allclose(out, out_exp, atol=1e-12)
        self.assertTrue(out.flags['C_CONTIGUOUS'])

        # 2D IFFT over last 2 axes of 4d array.
        fsignal = np.ones((1, 2, 5, 10), dtype=np.complex128)
        fsignal[0,1,:] = 10.
        out_exp = np.zeros((1, 2, 5, 10))
        out_exp[0,0,0,0] = 50
        out_exp[0,1,0,0] = 500
        out = fft.ifft(fsignal, axes=[-2, -1])
        np.testing.assert_allclose(out, out_exp, atol=1e-12)
        self.assertTrue(out.flags['C_CONTIGUOUS'])

        # 2D IFFT over last 2 axes of 4d non-contiguous array.
        fsignal = np.ones((1, 2, 5, 10), dtype=np.complex128)
        fsignal[0,1,:] = 10.
        tod = np.zeros((5, 10, 1, 2), dtype=np.complex128).transpose(2, 3, 0, 1)
        out_exp = np.zeros_like(tod)
        out_exp[0,0,0,0] = 50
        out_exp[0,1,0,0] = 500
        out = fft.ifft(fsignal, tod=tod, axes=[-2, -1])
        self.assertTrue(np.shares_memory(tod, out))
        np.testing.assert_allclose(out, out_exp, atol=1e-12)
        self.assertFalse(out.flags['C_CONTIGUOUS'])

        # 2D IFFT over middle 2 axes of 4d array.
        fsignal = np.ones((1, 5, 10, 2), dtype=np.complex128)
        fsignal[0,:,:,1] = 10.
        out_exp = np.zeros((1, 5, 10, 2))
        out_exp[0,0,0,0] = 50
        out_exp[0,0,0,1] = 500
        out = fft.ifft(fsignal, axes=[-3, -2])
        np.testing.assert_allclose(out, out_exp, atol=1e-12)
        self.assertTrue(out.flags['C_CONTIGUOUS'])

    def test_queb_rotmat_complex(self):
        # Tests that the rotmat respects fft symmetry constraints
        # for real maps -- ie, map2harm, harm2map produces
        # sensible round-trips. This version tests a map that's
        # complex in real-space
        ishapes = [(10, 10), (10, 11), (11, 10), (11, 11)]
        dtypes = [np.complex64, np.complex128]
        comps = [1, 2] # this is the comp we will set to non-zero

        for ishape, dtype, comp in itertools.product(ishapes, dtypes, comps):
            shape, wcs = enmap.geometry(pos=(0, 0), shape=ishape, res=0.01)

            # define some easy test input to evaluate for mixing
            input_complex_hmap = enmap.zeros((3, *shape), wcs, dtype=dtype)
            input_complex_hmap[comp] += 1. + 1.j
            output_complex_hmap = enmap.zeros((3, *shape), wcs, dtype=dtype)
            output_complex_hmap[comp] += 1. + 1.j

            # do a round-trip and check that we get what we expect.
            # the queb operations mix E and B and could lead to artifacts if
            # we're not careful
            if input_complex_hmap.real.dtype.itemsize == 8:
                atol = 1e-10
            elif input_complex_hmap.real.dtype.itemsize == 4:
                atol = 1e-5
            test_output_complex_hmap = enmap.map2harm(
                enmap.harm2map(input_complex_hmap, keep_imag=True)
                )
            assert np.allclose(
                test_output_complex_hmap, output_complex_hmap, rtol=0, atol=atol
                ), f'{ishape=}, {dtype=}, {comp=}'

    def test_queb_rotmat_real(self):
        # Tests that the rotmat respects fft symmetry constraints
        # for real maps -- ie, map2harm, harm2map produces
        # sensible round-trips. This version tests a map that's
        # real in real-space. This means that its DC must be real
        # and along even dimensions the nyquist freq must also be real
        ishapes = [(10, 10), (10, 11), (11, 10), (11, 11)]
        dtypes = [np.complex64, np.complex128]
        comps = [1, 2] # this is the comp we will set to non-zero
        np.random.seed(0)

        for ishape, dtype, comp in itertools.product(ishapes, dtypes, comps):
            shape, wcs = enmap.geometry(pos=(0, 0), shape=ishape, res=0.01)

            # define some easy test input to evaluate for mixing
            input_hmap = enmap.map2harm(enmap.rand_gauss((3,)+shape, wcs, dtype=dtype))
            # do a round-trip and check that we get what we expect.
            atol = 1e-10 if dtype == np.complex128 else 1e-5
            output_hmap = enmap.map2harm(enmap.harm2map(input_hmap))
            assert np.allclose(
                input_hmap, output_hmap, rtol=0, atol=atol), f'{ishape=}, {dtype=}, {comp=}'

    def test_extract(self):
        # Tests that extraction is sensible
        shape,wcs = enmap.geometry(pos=(0,0),shape=(500,500),res=0.01)
        imap = enmap.enmap(np.random.random(shape),wcs)
        smap = imap[200:300,200:300]
        sshape,swcs = smap.shape,smap.wcs
        smap2 = enmap.extract(imap,sshape,swcs)
        pixbox = enmap.pixbox_of(imap.wcs,sshape,swcs)
        # Do write and read test
        filename = "temporary_extract_map.fits" # NOT THREAD SAFE
        enmap.write_map(filename,imap)
        smap3 = enmap.read_map(filename,pixbox=pixbox)
        os.remove(filename)
        assert np.all(np.isclose(smap,smap2))
        assert np.all(np.isclose(smap,smap3))
        assert wcsutils.equal(smap.wcs,smap2.wcs)
        assert wcsutils.equal(smap.wcs,smap3.wcs)

    def test_fullsky_geometry(self):
        # Tests whether number of pixels and area of a full-sky 0.5 arcminute resolution map are correct
        print("Testing full sky geometry...")
        test_res_arcmin = 0.5
        shape,wcs = enmap.fullsky_geometry(res=np.deg2rad(test_res_arcmin/60.),proj='car')
        assert shape[0]==21600 and shape[1]==43200
        assert abs(enmap.area(shape,wcs) - 4*np.pi) < 1e-6

    def test_pixels(self):
        """Runs reference pixel and mean-square comparisons on extracts from randomly generated
        maps"""
        print("Testing reference pixels...")
        results,rname = get_extraction_test_results(TEST_DIR+"/tests.yml")
        cresults = pickle.load(open(DATA_PREFIX+"%s.pkl" % rname,'rb'))
        assert sorted(results.keys())==sorted(cresults.keys())
        for g in results.keys():
            assert sorted(results[g].keys())==sorted(cresults[g].keys())
            for s in results[g].keys():
                assert sorted(results[g][s].keys())==sorted(cresults[g][s].keys())
                for e in results[g][s].keys():
                    assert np.all(np.isclose(results[g][s][e],cresults[g][s][e]))


    def test_sim_slice(self):
        ps = powspec.read_spectrum(DATA_PREFIX+"test_scalCls.dat")[:1,:1]
        test_res_arcmin = 10.0
        lmax = 2000
        fact = 2.
        shape,wcs = enmap.fullsky_geometry(res=np.deg2rad(test_res_arcmin/60.),proj='car')
        omap = curvedsky.rand_map(shape, wcs, ps,lmax=lmax)
        ofunc = lambda ishape,iwcs: fact*enmap.extract(omap,ishape,iwcs)
        nmap = reproject.populate(shape,wcs,ofunc,maxpixy = 400,maxpixx = 400)
        assert np.all(np.isclose(nmap/omap,2.))


    # This is currently broken, but it's always been broken. For doubly-even
    # dimensions, the double-nyquist frequency entry has inconsistent sign.
    # Need to find a way to fix this.
    #def test_b_sign(self):
    #    """
    #    We generate a random IQU map with geometry such that cdelt[0]<0 and
    #    cdelt[1]>0.
    #    We transform this to TEB with map2harm and map2alm followed by 
    #    scalar harm2map and alm2map and use these as reference T,E,B maps.
    #    We flip the original map along the RA direction, Dec direction, or both.
    #    We transform this to TEB with map2harm and map2alm followed by 
    #    scalar harm2map and alm2map and use these as comparison T,E,B maps.
    #    We compare these maps.
    #    """
    #    cltt,clee,clbb,clte = powspec.read_spectrum(DATA_PREFIX+"cosmo2017_10K_acc3_lensedCls.dat",expand=None)

    #    ps_cmb = np.zeros((3,3,cltt.size))
    #    ps_cmb[0,0] = cltt
    #    ps_cmb[1,1] = clee
    #    ps_cmb[2,2] = clbb
    #    ps_cmb[1,0] = clte
    #    ps_cmb[0,1] = clte
    #    seed = 100

    #    # test all possible cdelt flips and shapes
    #    sels = (np.s_[...], np.s_[...,::-1], np.s_[...,::-1,:], np.s_[...,::-1,::-1])

    #    Ny, Nx = 1080, 2160 
    #    shapes = ((Ny, Nx), (Ny-1, Nx), (Ny, Nx-1), (Ny-1, Nx-1))
    #    for ishape, sel in itertools.product(shapes, sels):
    #        # Curved-sky
    #        lmax = 1000
    #        alm = curvedsky.rand_alm_healpy(ps_cmb,lmax=lmax,seed=seed)
    #        shape,iwcs = enmap.fullsky_geometry(shape=ishape)
    #        wcs = enmap.empty(shape,iwcs)[sel].wcs
    #        shape = (3,) + shape
    #        imap = curvedsky.alm2map(alm,enmap.empty(shape,wcs))
    #        oalm = curvedsky.map2alm(imap.copy(),lmax=lmax)
    #        rmap = curvedsky.alm2map(oalm,enmap.empty(shape,wcs),spin=0) # reference map

    #        imap2 = imap.copy()[sel]
    #        oalm = curvedsky.map2alm(imap2.copy(),lmax=lmax)
    #        rmap2 = curvedsky.alm2map(oalm,enmap.empty(shape,wcs),spin=0) # comparison map

    #        assert np.allclose(rmap[0],rmap2[0],atol=0,rtol=1e-7)
    #        assert np.allclose(rmap[1],rmap2[1],atol=0,rtol=1e-7)
    #        assert np.allclose(rmap[2],rmap2[2],atol=0,rtol=1e-7)
    #        
    #    Ny, Nx = 300, 300 
    #    shapes = ((Ny, Nx), (Ny-1, Nx), (Ny, Nx-1), (Ny-1, Nx-1))
    #    for ishape, sel in itertools.product(shapes, sels):
    #        # Flat-sky
    #        px = 2.0
    #        shape,iwcs = enmap.geometry(pos=(0,0),res=np.deg2rad(px/60.),shape=ishape)
    #        shape = (3,) + shape
    #        a = enmap.zeros(shape,iwcs)
    #        a = a[sel]
    #        wcs = a.wcs

    #        imap = enmap.rand_map(shape,wcs,ps_cmb,seed=seed)
    #        kmap = enmap.map2harm(imap.copy())
    #        rmap = enmap.harm2map(kmap,spin=0) # reference map

    #        imap = imap[sel]
    #        kmap = enmap.map2harm(imap.copy())
    #        rmap2 = enmap.harm2map(kmap,spin=0)[sel] # comparison map

    #        assert np.allclose(rmap[0],rmap2[0],atol=0,rtol=1e-7), f'{ishape=}, {sel=}'
    #        assert np.allclose(rmap[1],rmap2[1],atol=0,rtol=1e-7), f'{ishape=}, {sel=}'
    #        assert np.allclose(rmap[2],rmap2[2],atol=0,rtol=1e-7), f'{ishape=}, {sel=}'

    def test_plain_wcs(self):
        # Test area and box for a small Cartesian geometry
        shape,wcs = enmap.geometry(res=np.deg2rad(1./60.),shape=(600,600),pos=(0,0),proj='plain')
        box = np.rad2deg(enmap.box(shape,wcs))
        area = np.rad2deg(np.rad2deg(enmap.area(shape,wcs)))
        assert np.all(np.isclose(box,np.array([[-5,-5],[5,5]])))
        assert np.isclose(area,100.)

        # and for an artifical Cartesian geometry with area>4pi
        shape,wcs = enmap.geometry(res=np.deg2rad(10),shape=(100,100),pos=(0,0),proj='plain')
        box = np.rad2deg(enmap.box(shape,wcs))
        area = np.rad2deg(np.rad2deg(enmap.area(shape,wcs)))
        assert np.all(np.isclose(box,np.array([[-500,-500],[500,500]])))
        assert np.isclose(area,1000000)


    def test_pospix(self):
        # Posmap separable and non-separable on CAR
        for res in [6,12,24]:
            shape,wcs = enmap.fullsky_geometry(res=np.deg2rad(res/60.),proj='car')
            posmap1 = enmap.posmap(shape,wcs)
            posmap2 = enmap.posmap(shape,wcs,separable=True)
            assert np.all(np.isclose(posmap1,posmap2))

        # Pixmap plain
        pres = 0.5
        shape,wcs = enmap.geometry(pos=(0,0),shape=(30,30),res=pres*u.degree,proj='plain')
        yp,xp = enmap.pixshapemap(shape,wcs)
        assert np.all(np.isclose(yp,pres*u.degree))
        assert np.all(np.isclose(xp,pres*u.degree))
        yp,xp = enmap.pixshape(shape,wcs)
        parea = enmap.pixsize(shape,wcs)
        assert np.isclose(parea,(pres*u.degree)**2)
        assert np.isclose(yp,pres*u.degree)
        assert np.isclose(xp,pres*u.degree)
        pmap = enmap.pixsizemap(shape,wcs)
        assert np.all(np.isclose(pmap,(pres*u.degree)**2))

        # Pixmap CAR
        pres = 0.1
        dec_cut = 89.5 # pixsizemap is not accurate near the poles currently
        shape,wcs = enmap.band_geometry(dec_cut=dec_cut*u.degree,res=pres*u.degree,proj='car')
        # Current slow and general but inaccurate near the poles implementation
        pmap = enmap.pixsizemap(shape,wcs)
        # Fast CAR-specific pixsizemap implementation
        dra, ddec = wcs.wcs.cdelt*u.degree
        dec = enmap.posmap([shape[-2],1],wcs)[0,:,0]
        area = np.abs(dra*(np.sin(np.minimum(np.pi/2.,dec+ddec/2))-np.sin(np.maximum(-np.pi/2.,dec-ddec/2))))
        Nx = shape[-1]
        pmap2 = enmap.ndmap(area[...,None].repeat(Nx,axis=-1),wcs)
        assert np.all(np.isclose(pmap,pmap2))
        

    def test_project_nn(self):
        shape,wcs = enmap.fullsky_geometry(res=np.deg2rad(12/60.),proj='car')
        shape2,wcs2 = enmap.fullsky_geometry(res=np.deg2rad(6/60.),proj='car')
        shape3,wcs3 = enmap.fullsky_geometry(res=np.deg2rad(24/60.),proj='car')
        imap = enmap.ones(shape,wcs)
        omap2 = enmap.project(imap,shape2,wcs2,order=0,border='wrap')
        omap3 = enmap.project(imap,shape3,wcs3,order=0,border='wrap')
        assert np.all(np.isclose(omap2,1))
        assert np.all(np.isclose(omap3,1))

    def test_wcsunequal(self):
        shape1,wcs1 = enmap.geometry(pos=(0,0),shape=(100,100),res=1*u.arcmin,proj='car')
        shape1,wcs2 = enmap.geometry(pos=(0,0),shape=(100,100),res=1*u.arcmin,proj='cea')
        shape1,wcs3 = enmap.geometry(pos=(10,10),shape=(100,100),res=1*u.arcmin,proj='car')
        shape1,wcs4 = enmap.geometry(pos=(0,0),shape=(100,100),res=2*u.arcmin,proj='car')
        assert not(wcsutils.equal(wcs1,wcs2))
        assert not(wcsutils.equal(wcs1,wcs3))
        assert not(wcsutils.equal(wcs1,wcs4))
        
        
    def test_scale(self):
        # Test (with a plain geometry) that scale_geometry
        # will result in geometries with the same bounding box
        # but different area pixel
        pres = 0.5
        ufact = 2
        dfact = 0.5
        shape,wcs = enmap.geometry(pos=(0,0),shape=(30,30),res=pres*u.arcmin,proj='plain')
        ushape,uwcs = enmap.scale_geometry(shape,wcs,ufact)
        dshape,dwcs = enmap.scale_geometry(shape,wcs,dfact)
        box = enmap.box(shape,wcs)
        ubox = enmap.box(ushape,uwcs)
        dbox = enmap.box(dshape,dwcs)
        parea = enmap.pixsize(shape,wcs)
        uparea = enmap.pixsize(ushape,uwcs)
        dparea = enmap.pixsize(dshape,dwcs)
        assert np.all(np.isclose(box,ubox))
        assert np.all(np.isclose(box,dbox))
        assert np.isclose(parea/(ufact**2),uparea)
        assert np.isclose(parea/(dfact**2),dparea)

    def test_prepare_alm_mmax(self):
        # Check if mmax is correctly handled by prepare_alm.

        # Create lmax=mmax=3 alm array and corresponding alm_info.
        lmax = 3
        nalm = 10  # Triangular alm array of lmax=3 has 10 elements.
        alm_in = np.arange(nalm, dtype=np.complex128)
        ainfo_in = curvedsky.alm_info(
            lmax=3, mmax=3, nalm=nalm, stride=1, layout="triangular")

        # Case 1: provide only alm.
        alm_out, ainfo_out = curvedsky.prepare_alm(alm=alm_in, ainfo=None)

        np.testing.assert_array_almost_equal(alm_out, alm_in)
        self.assertEqual(ainfo_out.lmax, ainfo_in.lmax)
        self.assertEqual(ainfo_out.mmax, ainfo_in.mmax)
        self.assertEqual(ainfo_out.nelem, ainfo_in.nelem)

        # Case 2: provide only alm_info.
        alm_out, ainfo_out = curvedsky.prepare_alm(alm=None, ainfo=ainfo_in)
        # Expect zero array.
        np.testing.assert_array_almost_equal(alm_out, alm_in * 0)
        self.assertEqual(ainfo_out.lmax, ainfo_in.lmax)
        self.assertEqual(ainfo_out.mmax, ainfo_in.mmax)
        self.assertEqual(ainfo_out.nelem, ainfo_in.nelem)

        # Case 3: provide alm and alm_info
        alm_out, ainfo_out = curvedsky.prepare_alm(alm=alm_in, ainfo=ainfo_in)

        np.testing.assert_array_almost_equal(alm_out, alm_in)
        self.assertEqual(ainfo_out.lmax, ainfo_in.lmax)
        self.assertEqual(ainfo_out.mmax, ainfo_in.mmax)
        self.assertEqual(ainfo_out.nelem, ainfo_in.nelem)

        # Case 4: provide only alm with lmax=3 and mmax=1.
        # This should currently fail.
        nalm = 7
        alm_in = np.arange(7, dtype=np.complex128)
        self.assertRaises(AssertionError, curvedsky.prepare_alm,
                          **dict(alm=alm_in, ainfo=None, lmax=lmax))

        # Case 5: provide only alm_info with lmax=3 and mmax=1.
        nalm = 7
        ainfo_in = curvedsky.alm_info(
            lmax=3, mmax=1, nalm=nalm, stride=1, layout="triangular")
        alm_exp = np.zeros(7, dtype=np.complex128)
        alm_out, ainfo_out = curvedsky.prepare_alm(alm=None, ainfo=ainfo_in)

        np.testing.assert_array_almost_equal(alm_out, alm_exp)
        self.assertEqual(ainfo_out.lmax, ainfo_in.lmax)
        self.assertEqual(ainfo_out.mmax, ainfo_in.mmax)
        self.assertEqual(ainfo_out.nelem, ainfo_in.nelem)

        # Case 6: provide both alm and alm_info with lmax=3 and mmax=1.
        # This should be allowed.
        nalm = 7
        ainfo_in = curvedsky.alm_info(
            lmax=3, mmax=1, nalm=nalm, stride=1, layout="triangular")
        alm_in = np.arange(7, dtype=np.complex128)
        alm_out, ainfo_out = curvedsky.prepare_alm(alm=alm_in, ainfo=ainfo_in)

        np.testing.assert_array_almost_equal(alm_out, alm_in)
        self.assertEqual(ainfo_out.lmax, ainfo_in.lmax)
        self.assertEqual(ainfo_out.mmax, ainfo_in.mmax)
        self.assertEqual(ainfo_out.nelem, ainfo_in.nelem)


    def test_lens_alms(self):
        # We generate phi alms and convert them to kappa and back
        lmax = 100
        ps = np.zeros(lmax+1)
        ls = np.arange(lmax+1)
        ps[ls>=2] = 1./ls[ls>=2]
        phi_alm = curvedsky.rand_alm(ps,lmax=lmax)
        kappa_alm = lensing.phi_to_kappa(phi_alm)
        phi_alm2 = lensing.kappa_to_phi(kappa_alm)
        np.testing.assert_array_almost_equal(phi_alm, phi_alm2)

    def test_downgrade(self):
        shape,wcs = enmap.geometry(pos=(0,0),shape=(100,100),res=0.01)
        imap = enmap.ones(shape,wcs)
        for dfact in [None,1]:
            omap = enmap.downgrade(imap,dfact)
            np.testing.assert_equal(imap,omap)

        dfact = 2
        omap = enmap.downgrade(imap,dfact,op=np.sum)
        np.testing.assert_equal(omap,np.ones(enmap.scale_geometry(shape,wcs,1./dfact)[0])*4)

        
    def test_almxfl(self):
        # We try to filter alms of shape (nalms,) and (ncomp,nalms) with
        # a filter of shape (nells,)
        lmax = 30
        ells = np.arange(lmax+1)
        nells = ells.size
        for ncomp in range(4):
            if ncomp==0:
                fl = np.ones((nells,))
                ps = np.zeros((nells,))
                ps[ells>1] = 1./ells[ells>1]
            else:
                fl = np.ones((nells))
                ps = np.zeros((ncomp,ncomp,nells))
                for i in range(ncomp):
                    ps[i,i][ells>1] = 1./ells[ells>1]
            ialm = curvedsky.rand_alm(ps,lmax=lmax)
            oalm = curvedsky.almxfl(ialm,fl)
            np.testing.assert_array_almost_equal(ialm, oalm)

    def test_alm2map_2d_roundtrip(self):
        # Test curvedsky's alm2map/map2alm.
        lmax = 30
        ainfo = curvedsky.alm_info(lmax)

        nrings = lmax + 2
        nphi = 2 * lmax + 1
        shape, wcs = enmap.fullsky_geometry(shape=(nrings,nphi))

        # Test different input shapes and dtypes.
        # Case 1a: 1d double precision.
        spin = 0
        alm = np.zeros((ainfo.nelem), dtype=np.complex128)
        i   = ainfo.lm2ind(lmax,lmax)
        alm[i] = 1. + 1.j

        omap = enmap.zeros(shape, wcs, np.float64)
        curvedsky.alm2map(alm, omap, spin=spin)
        alm_out = curvedsky.map2alm(omap, spin=spin, ainfo=ainfo)

        np.testing.assert_array_almost_equal(alm_out, alm)

        # Case 1b: 1d single precision.
        spin = 0
        alm = np.zeros((ainfo.nelem), dtype=np.complex64)
        alm[i] = 1. + 1.j

        omap = enmap.zeros(shape, wcs, np.float32)
        curvedsky.alm2map(alm, omap, spin=spin)
        alm_out = curvedsky.map2alm(omap, spin=spin, ainfo=ainfo)
        np.testing.assert_array_almost_equal(alm_out, alm)

        # Case 2a: 2d double precision.
        spin = 1
        nspin = 2
        alm = np.zeros((nspin, ainfo.nelem), dtype=np.complex128)
        alm[0,i] = 1. + 1.j
        alm[1,i] = 2. - 2.j

        omap = enmap.zeros((nspin,)+shape, wcs, np.float64)
        curvedsky.alm2map(alm, omap, spin=spin)
        alm_out = curvedsky.map2alm(omap, spin=spin, ainfo=ainfo)
        np.testing.assert_array_almost_equal(alm_out, alm)

        # Case 2b: 2d single precision.
        spin = 1
        nspin = 2
        alm = np.zeros((nspin, ainfo.nelem), dtype=np.complex64)
        alm[0,i] = 1. + 1.j
        alm[1,i] = 2. - 2.j

        omap = enmap.zeros((nspin,)+shape, wcs, np.float32)
        curvedsky.alm2map(alm, omap, spin=spin)
        alm_out = curvedsky.map2alm(omap, spin=spin, ainfo=ainfo)
        np.testing.assert_array_almost_equal(alm_out, alm)

        # Case 3a: 3d double precision.
        spin = 1
        nspin = 2
        ntrans = 3
        alm = np.zeros((ntrans, nspin, ainfo.nelem), dtype=np.complex128)
        alm[0,0,i] = 1. + 1.j
        alm[0,1,i] = 2. - 2.j
        alm[1,0,i] = 3. + 3.j
        alm[1,1,i] = 4. - 4.j
        alm[2,0,i] = 5. + 5.j
        alm[2,1,i] = 6. - 6.j

        omap = enmap.zeros((ntrans,nspin)+shape, wcs, np.float64)
        curvedsky.alm2map(alm, omap, spin=spin)
        alm_out = curvedsky.map2alm(omap, spin=spin, ainfo=ainfo)
        np.testing.assert_array_almost_equal(alm_out, alm)

        # Case 3b: 3d single precision.
        spin = 1
        nspin = 2
        ntrans = 3
        alm = np.zeros((ntrans, nspin, ainfo.nelem), dtype=np.complex64)
        alm[0,0,i] = 1. + 1.j
        alm[0,1,i] = 2. - 2.j
        alm[1,0,i] = 3. + 3.j
        alm[1,1,i] = 4. - 4.j
        alm[2,0,i] = 5. + 5.j
        alm[2,1,i] = 6. - 6.j

        omap = enmap.zeros((ntrans,nspin)+shape, wcs, np.float32)
        curvedsky.alm2map(alm, omap, spin=spin)
        alm_out = curvedsky.map2alm(omap, spin=spin, ainfo=ainfo)
        np.testing.assert_array_almost_equal(alm_out, alm)

    def test_alm2map_healpix_roundtrip(self):
        # Test curvedsky's alm2map/map2alm.
        nside = 2
        lmax  = nside*2
        nside = lmax//2
        ainfo = curvedsky.alm_info(lmax)
        npix  = 12*nside**2
        # 7 iterations needed to reach 6 digits of
        # precision. This is more than the 3 default
        # in healpy and the 0 default in pixell
        niter = 7

        for dtype in [np.float64, np.float32]:
            ctype = utils.complex_dtype(dtype)

            # Case 1: 1d
            spin = 0
            alm = np.zeros((ainfo.nelem), dtype=ctype)
            i   = ainfo.lm2ind(lmax,lmax)
            alm[i] = 1. + 1.j

            omap = np.zeros(npix, dtype)
            curvedsky.alm2map_healpix(alm, omap, spin=spin)
            alm_out = curvedsky.map2alm_healpix(omap, spin=spin, ainfo=ainfo, niter=niter)

            np.testing.assert_array_almost_equal(alm_out, alm)

            # Case 2: 2d
            spin = 1
            nspin = 2
            alm = np.zeros((nspin, ainfo.nelem), dtype=ctype)
            alm[0,i] = 1. + 1.j
            alm[1,i] = 2. - 2.j

            omap = np.zeros((nspin,npix), dtype)
            curvedsky.alm2map_healpix(alm, omap, spin=spin)
            alm_out = curvedsky.map2alm_healpix(omap, spin=spin, ainfo=ainfo, niter=niter)
            np.testing.assert_array_almost_equal(alm_out, alm)

            # Case 3: 3d
            spin = 1
            nspin = 2
            ntrans = 3
            alm = np.zeros((ntrans, nspin, ainfo.nelem), dtype=ctype)
            alm[0,0,i] = 1. + 1.j
            alm[0,1,i] = 2. - 2.j
            alm[1,0,i] = 3. + 3.j
            alm[1,1,i] = 4. - 4.j
            alm[2,0,i] = 5. + 5.j
            alm[2,1,i] = 6. - 6.j

            omap = np.zeros((ntrans,nspin,npix), dtype)
            curvedsky.alm2map_healpix(alm, omap, spin=spin)
            alm_out = curvedsky.map2alm_healpix(omap, spin=spin, ainfo=ainfo, niter=niter)
            np.testing.assert_array_almost_equal(alm_out, alm)

    # MM: Re-enabled 09/17/2024
    # --Disabled for now because the version of ducc currently on pypi
    # has an adjointness bug. It's fixed in the ducc git repo.--
    def test_adjointness(self):
       # This tests if alm2map_adjoint is the adjoint of alm2map,
       # and if map2alm_adjoint is the adjoint of map2alm.
       # (This doesn't test if they're correct, just that they're
       # consistent with each other). This test is a bit slow, taking
       # 5 s or so. It would be much faster if we dropped the ncomp=3 case.
       for dtype in [np.float32, np.float64]:
           for ncomp in [1,3]:
               # Define our geometries
               geos = []
               res  = 30*utils.degree
               shape, wcs = enmap.fullsky_geometry(res=res, variant="fejer1")
               geos.append(("fullsky_fejer1", shape, wcs))

               shape, wcs = enmap.fullsky_geometry(res=res, variant="cc")
               geos.append(("fullsky_cc", shape, wcs))
               lmax = shape[-2]-2

               shape, wcs = enmap.Geometry(shape, wcs)[3:-3,3:-3]
               geos.append(("patch_cc", shape, wcs))

               wcs = wcs.deepcopy()
               wcs.wcs.crpix += 0.123
               geos.append(("patch_gen_cyl", shape, wcs))

               shape, wcs = enmap.geometry(np.array([[-45,45],[45,-45]])*utils.degree, res=res, proj="tan")
               geos.append(("patch_tan", shape, wcs))

               for gi, (name, shape, wcs) in enumerate(geos):
                   mat1  = alm_bash(curvedsky.alm2map,         shape, wcs, ncomp, lmax, dtype)
                   mat2  = map_bash(curvedsky.alm2map_adjoint, shape, wcs, ncomp, lmax, dtype)
                   np.testing.assert_array_almost_equal(mat1, mat2)
                   mat1 = map_bash(curvedsky.map2alm,         shape, wcs, ncomp, lmax, dtype)
                   mat2 = alm_bash(curvedsky.map2alm_adjoint, shape, wcs, ncomp, lmax, dtype)
                   np.testing.assert_array_almost_equal(mat1, mat2)


    #def test_sharp_alm2map_der1(self):
    #            
    #    # Test the wrapper around libsharps alm2map_der1.
    #    lmax = 3
    #    ainfo = sharp.alm_info(lmax)

    #    nrings = lmax + 1
    #    nphi = 2 * lmax + 1
    #    minfo = sharp.map_info_gauss_legendre(nrings, nphi)

    #    sht = sharp.sht(minfo, ainfo)

    #    # Test different input shapes and dtypes.
    #    # Case 1a: 1d double precision.
    #    alm = np.zeros((ainfo.nelem), dtype=np.complex128)
    #    alm[4] = 1. + 1.j

    #    omap = sht.alm2map_der1(alm)
    #    # Compare to expected value by doing spin 1 transform
    #    # on sqrt(ell (ell + 1)) alm.
    #    alm_spin = np.zeros((2, ainfo.nelem), dtype=np.complex128)
    #    alm_spin[0] = alm * np.sqrt(2)
    #    omap_exp = sht.alm2map(alm_spin, spin=1)

    #    np.testing.assert_array_almost_equal(omap, omap_exp)

    #    # Case 1b: 1d single precision.
    #    alm = np.zeros((ainfo.nelem), dtype=np.complex64)
    #    alm[4] = 1. + 1.j

    #    omap = sht.alm2map_der1(alm)
    #    # Compare to expected value by doing spin 1 transform
    #    # on sqrt(ell (ell + 1)) alm.
    #    alm_spin = np.zeros((2, ainfo.nelem), dtype=np.complex64)
    #    alm_spin[0] = alm * np.sqrt(2)
    #    omap_exp = sht.alm2map(alm_spin, spin=1)

    #    np.testing.assert_array_almost_equal(omap, omap_exp)

    #    # Case 2a: 2d double precision.
    #    ntrans = 3
    #    alm = np.zeros((ntrans, ainfo.nelem), dtype=np.complex128)
    #    alm[0,4] = 1. + 1.j
    #    alm[1,4] = 2. + 2.j
    #    alm[2,4] = 3. + 3.j

    #    omap = sht.alm2map_der1(alm)
    #    # Compare to expected value by doing spin 1 transform
    #    # on sqrt(ell (ell + 1)) alm.
    #    alm_spin = np.zeros((ntrans, 2, ainfo.nelem), dtype=np.complex128)
    #    alm_spin[0,0] = alm[0] * np.sqrt(2)
    #    alm_spin[1,0] = alm[1] * np.sqrt(2)
    #    alm_spin[2,0] = alm[2] * np.sqrt(2)
    #    omap_exp = sht.alm2map(alm_spin, spin=1)

    #    np.testing.assert_array_almost_equal(omap, omap_exp)

    #    # Case 2b: 2d single precision.
    #    ntrans = 3
    #    alm = np.zeros((ntrans, ainfo.nelem), dtype=np.complex64)
    #    alm[0,4] = 1. + 1.j
    #    alm[1,4] = 2. + 2.j
    #    alm[2,4] = 3. + 3.j

    #    omap = sht.alm2map_der1(alm)
    #    # Compare to expected value by doing spin 1 transform
    #    # on sqrt(ell (ell + 1)) alm.
    #    alm_spin = np.zeros((ntrans, 2, ainfo.nelem), dtype=np.complex64)
    #    alm_spin[0,0] = alm[0] * np.sqrt(2)
    #    alm_spin[1,0] = alm[1] * np.sqrt(2)
    #    alm_spin[2,0] = alm[2] * np.sqrt(2)
    #    omap_exp = sht.alm2map(alm_spin, spin=1)

    #    np.testing.assert_array_almost_equal(omap, omap_exp)

    def test_thumbnails(self):
        print("Testing thumbnails...")

        # Make a geometry far away from the equator
        dec_min = 70 * u.degree
        dec_max = 80 * u.degree
        res = 0.5 * u.arcmin
        shape,wcs = enmap.band_geometry((dec_min,dec_max),res=res)

        # Create a set of point source positions separated by
        # 2 degrees but with 1 column wrapping around the RA
        # direction
        width = 120 * u.arcmin
        Ny = int((dec_max-dec_min)/(width))
        Nx = int((2*np.pi/(width)))
        pys = np.linspace(0,shape[0],Ny)[1:-1]
        pxs = np.linspace(0,shape[1],Nx)[:-1]
        Ny = len(pys)
        Nx = len(pxs)
        xx,yy = np.meshgrid(pxs,pys)
        xx = xx.reshape(-1)
        yy = yy.reshape(-1)
        ps = np.vstack((yy,xx))
        decs,ras = enmap.pix2sky(shape,wcs,ps)
        
        # Simulate these sources with unit peak value and 2.5 arcmin FWHM
        N = ps.shape[1]
        srcs = np.zeros((N,3))
        srcs[:,0] = decs
        srcs[:,1] = ras
        srcs[:,2] = ras*0 + 1
        sigma = 2.5 * u.fwhm * u.arcmin
        omap = pointsrcs.sim_srcs(shape,wcs,srcs,beam=sigma)

        # Reproject thumbnails centered on the sources
        # with gnomonic/tangent projection
        proj = "tan"
        r = 10*u.arcmin
        ret = reproject.thumbnails(omap, srcs[:,:2], r=r, res=res, proj=proj,
            apod=2*u.arcmin, order=3, oversample=4, pixwin=False)

        # Create a reference source at the equator to compare this against
        ishape,iwcs = enmap.geometry(shape=ret.shape,res=res,pos=(0,0),proj=proj)
        imodrmap = enmap.modrmap(ishape,iwcs)
        model = np.exp(-imodrmap**2./2./sigma**2.)

        # Make sure all thumbnails agree with the reference at the
        # sub-percent level
        for i in range(ret.shape[0]):
            diff = ret[i] - model
            assert np.all(np.isclose(diff,0,atol=1e-3))

    def test_tilemap(self):
        shape, wcs = enmap.fullsky_geometry(30*utils.degree, variant="CC")
        assert shape == (7,12)
        geo  = tilemap.geometry((3,)+shape, wcs, tile_shape=(2,2))
        assert len(geo.active) == 0
        assert np.all(geo.lookup<0)
        assert geo.ntile   == 24
        assert geo.nactive == 0
        assert geo.tile_shape == (2,2)
        assert geo.grid_shape == (4,6)
        assert tuple(geo.tile_shapes[ 0]) == (2,2)
        assert tuple(geo.tile_shapes[ 5]) == (2,2)
        assert tuple(geo.tile_shapes[18]) == (1,2)
        assert tuple(geo.tile_shapes[23]) == (1,2)
        assert geo.ind2grid(7) == (1,1)
        assert geo.grid2ind(1,1) == 7
        geo = geo.copy(active=[1])
        assert geo.nactive == 1
        assert np.sum(geo.lookup>=0) == 1
        assert geo.active[0] == 1
        assert geo.lookup[1] == 0
        geo2 = geo.copy(active=[0,1,2])
        assert geo.nactive == 1
        assert geo2.nactive == 3
        assert geo.compatible(geo) == 2
        assert geo.compatible(geo2) == 1
        geo3 = tilemap.geometry((3,)+shape, wcs, tile_shape=(2,3))
        assert geo.compatible(geo3) == 0
        del geo2, geo3
        m1  = tilemap.zeros(geo.copy(active=[1,2]))
        m2  = tilemap.zeros(geo.copy(active=[2,3,4]))
        m3  = tilemap.zeros(geo.copy(active=[2]))
        for a, i in enumerate(m1.geometry.active): m1.active_tiles[a] = i
        for a, i in enumerate(m2.geometry.active): m2.active_tiles[a] = i*10
        for a, i in enumerate(m3.geometry.active): m3.active_tiles[a] = i*100
        assert m1[0,0] == 1
        assert np.all(m1.tiles[1] == m1.active_tiles[0])
        m12 = m1+m2
        m21 = m2+m1
        assert(m12.nactive == 4)
        assert(m21.nactive == 4)
        assert(np.all(m12.tiles[1] == 1))
        assert(np.all(m21.tiles[1] == 1))
        assert(np.all(m12.tiles[2] == 22))
        assert(np.all(m21.tiles[2] == 22))
        assert(sorted(m12.geometry.active)==sorted(m21.geometry.active))
        m1 += m3
        assert np.all(m1.tiles[2] == 202)
        with self.assertRaises(ValueError): m3 += m1
        m1[:] = 0
        m1c   = np.cos(m1)
        assert m1c.geometry.nactive == 2
        assert np.allclose(m1c, 1)

    def test_interpol_1d(self):
        dtype = np.float64
        n     = 10
        data  = np.sin(2*np.pi*3*np.arange(n)/n)
        inds  = np.array([0.0,0.1,0.51,0.9,1.0])
        # Nearest neighbor interpolation
        vals  = utils.interpol(data, inds[None], mode="spline", order=0)
        assert np.allclose(vals, data[utils.nint(inds)])
        # Linear interpolation. This check assumes that inds is [0:1]!
        vals  = utils.interpol(data, inds[None], mode="spline", order=1)
        assert np.allclose(vals, data[0]*(1-inds)+data[1]*inds)
        # Cubic spline interpolation. No simple formula here, so hardcode
        vals  = utils.interpol(data, inds[None], mode="spline", order=3)
        assert np.allclose(vals, [-1.39824833112681842e-12, 1.16478779103952171e-01, 6.66141591529904153e-01, 9.62884886419913988e-01, 9.51056516295153753e-01])
        # Nufft interpolation
        vals  = utils.interpol(data, inds[None], mode="fourier")
        targ  = np.sin(2*np.pi*3*inds/n)
        assert np.allclose(vals, targ)

    def test_interpol_map(self):
        # Make a small map to interpolate. We will simply use the angular distance
        # from the center, which avoids any non-periodicity issues
        shape, wcs = enmap.geometry([[-1,1],[1,-1]], res=0.1, deg=True, proj="car")
        # Make test cases where the exact answer is known. Bilinear is easy for a linear field,
        # and NN is easy anywayre
        dtype = np.float64
        ypix  = np.arange(shape[-2], dtype=dtype)
        xpix  = np.arange(shape[-1], dtype=dtype)
        opix  = np.array([[10,10.2],[10,12.7]])
        # NN and bilinear
        def f(y,x): return 2*y-x
        data  = enmap.enmap(f(ypix[:,None],xpix[None,:]),wcs)
        vals  = data.at(opix, unit="pix", mode="spline", order=0)
        targ  = data[tuple(utils.nint(opix))]
        assert np.allclose(vals, targ)
        vals  = data.at(opix, unit="pix", mode="spline", order=1)
        targ  = f(*opix)
        assert np.allclose(vals, targ)
        # Bicubic is exact up to 3rd order, but only if we ignore
        # boundary conditions. I'm not sure how to calculate the
        # expected answer here, so just hardcode it
        def f(y,x): return y**3+2*y**2+3*y-2*x**3-3*x**2-4*x+x*y
        data  = enmap.enmap(f(ypix[:,None],xpix[None,:]),wcs)
        vals  = data.at(opix, unit="pix", mode="spline", order=3)
        targ  = np.array([-1.01000000000000000e+03, -3.20214384455599429e+03])
        assert np.allclose(vals, targ)
        # Fourier is simple. All fourier modes up to Nyquist are exact.
        # Nufft adds some inaccuracy, but it's tiny
        def f(y,x): return np.sin(2*np.pi*3*y/shape[-2])+np.cos(2*np.pi*5*x/shape[-1])
        data  = enmap.enmap(f(ypix[:,None],xpix[None,:]),wcs)
        vals  = data.at(opix, unit="pix", mode="fourier")
        targ  = f(*opix)
        assert np.allclose(vals, targ)

    def test_interpol_map_Nd(self):
        # Test that multidimensional interpolation works
        shape, wcs = enmap.geometry([[-1,1],[1,-1]], res=0.1, deg=True, proj="car")
        dtype = np.float64
        ypix  = np.arange(shape[-2], dtype=dtype)
        xpix  = np.arange(shape[-1], dtype=dtype)
        def f(y,x): return np.sin(2*np.pi*3*y/shape[-2])+np.cos(2*np.pi*5*x/shape[-1])
        data  = (1+np.arange(12)).reshape(3,4)[:,:,None,None]*enmap.enmap(f(ypix[:,None],xpix[None,:]),wcs)
        opix  = np.array([[10,10.2],[10,12.7]])
        vals  = data.at(opix, unit="pix")
        assert vals.shape == (3,4)+opix.shape[1:]
        targ  = vals*0
        for I in utils.nditer(data.shape[:-2]):
            targ[I] = data[I].at(opix, unit="pix")
        assert np.allclose(vals, targ)

    def test_interpolator(self):
        # Test that lookups using an interpolator work, and are fast.
        # Use a decently large array to be sure we can measure the
        # speed difference.
        shape = (1000,1000)
        dtype = np.float64
        ypix  = np.arange(shape[-2], dtype=dtype)
        xpix  = np.arange(shape[-1], dtype=dtype)
        opix  = np.array([[10,10.2],[10,12.7]])
        def f(y,x): return np.sin(2*np.pi*3*y/shape[-2])+np.cos(2*np.pi*5*x/shape[-1])
        data  = f(ypix[:,None],xpix)
        cases = [("spline",3),("fourier",None)]
        for ci, (mode,order) in enumerate(cases):
            t1 = time.time()
            vals_raw = utils.interpol(data, opix, mode=mode, order=order)
            t2 = time.time()
            ip = utils.interpolator(data, mode=mode, order=order)
            t3 = time.time()
            vals_ip  = ip(opix)
            t4 = time.time()
            assert np.allclose(vals_ip, vals_raw)
            time_full  = t2-t1
            time_eval  = t4-t3
            # Should be much faster than 10%, but timing is unreliable, especially in github actions
            assert time_eval < time_full / 10
            # Also test that we can pass the interpolator as an argument
            vals_ip2 = utils.interpol(data, opix, mode=mode, order=order, ip=ip)
            assert np.allclose(vals_ip2, vals_raw)



cdef extern from "distances_core.h":
	ctypedef long long int inum
	ctypedef unsigned char uint8_t
	void distance_from_points_simple(inum npix, double * posmap, inum npoint, double * points, double * dists, int * area)
	void distance_from_points_simple_separable(inum ny, inum nx, double * ypos, double * xpos, inum npoint, double * points, double * dists, int * area)
	void distance_from_points_bubble(int ny, int nx, double * posmap, inum npoint, double * point_pos, int * point_pix, double rmax, double * dists, int * domains)
	void distance_from_points_bubble_separable(int ny, int nx, double * ypos, double * xpos, inum npoint, double * point_pos, int * point_pix, double rmax, double * dists, int * domains)
	inum find_edges(inum ny, inum nx, uint8_t * mask, inum ** edges)
	inum find_edges_labeled(inum ny, inum nx, int * labels, inum ** edges)

	void distance_from_points_cellgrid(int ny, int nx, double * ypos, double * xpos, inum npoint, double * point_pos, int * point_pix, int bsize_y, int bsize_x, double rmax, double dr, int separable, double * dists, int * domains)

	# Healpix stuff below
	ctypedef struct healpix_info:
		int nside, ny
		inum npix, ncap
		int * nx
		inum * off
		int  * shift
		double *ra0
		double *dec
		double *cos_dec
		double *sin_dec

	healpix_info * build_healpix_info(int nside)
	void free_healpix_info(healpix_info * geo)
	void unravel_healpix(healpix_info * geo, inum npoint, inum * pix1d, int * pix2d)
	void ravel_healpix(healpix_info * geo, inum npoint, int * pix2d, inum * pix1d)
	void get_healpix_neighs(healpix_info * geo, int y, int x, int * oy, int * ox)
	inum find_edges_healpix(healpix_info * geo, uint8_t * mask, int ** edges)
	inum find_edges_labeled_healpix(healpix_info * geo, int * labels, int ** edges)
	void distance_from_points_bubble_healpix(healpix_info * geo, inum npoint, double * point_pos, int * point_pix, double rmax, double * dists, int * domains)
	void distance_from_points_heap_healpix(healpix_info * geo, inum npoint, double * point_pos, int * point_pix, double rmax, double * dists, int * domains)



"""This is a module for analysing sky maps: stuff like matched filtering, feature
detection, pixel-space likelihoods etc.


Example usage of matched filter functions:

import numpy as np
from pixell import enmap, utils, uharm, analysis, curvedsky

np.random.seed(1)

# 0. Set up our geometry, a 100 uK peak point source and a 1.4 arcmin beam
pos        = [0,0]
shape, wcs = enmap.geometry(np.array([[-2,2],[2,-2]])*utils.degree, res=0.5*utils.arcmin)
pixarea    = enmap.pixsizemap(shape, wcs)
bsigma     = 1.4*utils.fwhm*utils.arcmin
signal     = 100*np.exp(-0.5*enmap.modrmap(shape, wcs, pos)**2/bsigma**2)
uht        = uharm.UHT(shape, wcs)
beam       = np.exp(-0.5*uht.l**2*bsigma**2)
fconv      = utils.dplanck(150e9, utils.T_cmb)/1e3 # uK -> mJy/sr

# 1. Matched filter for 10 uK' white noise
ivar       = 10**-2*pixarea/utils.arcmin**2
noise      = enmap.rand_gauss(shape, wcs)/ivar**0.5
map        = signal # + noise  # uncomment to actually add noise
# fconv is used to convert map and ivar from uK to mJy/sr. That way
# our output flux will be in mJy instead of the weird unit uK*sr.
rho, kappa = analysis.matched_filter_white(map*fconv, beam, ivar/fconv**2, uht)
flux  = rho.at(pos)/kappa.at(pos)
dflux = kappa.at(pos)**-0.5
print("%-15s %8.3f %8.3f %8.3f" % ("white", flux, dflux, flux/dflux))
# white              7.487    0.711   10.537

# 2. same, but with a noise power spectrum instead of a noise map. 10 uK'
# white noise has a flat noise spectrum of 10**2 * arcmin**2 = 0.46e-6
iN = 10**-2/utils.arcmin**2
rho, kappa = analysis.matched_filter_constcov(map*fconv, beam, iN/fconv**2, uht)
flux  = rho.at(pos)/kappa # kappa just a number in this case
dflux = kappa**-0.5
print("%-15s %8.3f %8.3f %8.3f" % ("constcov white", flux, dflux, flux/dflux))
# constcov white     7.486    0.711   10.534

# 3. uniform white noise, but with support for noise spectrum and position-dependence.
#    The noise units are only in ivar, so the noise spectrum is just a dimensionless 1.
iN = 1
rho, kappa = analysis.matched_filter_constcorr_lowcorr(map*fconv, beam, ivar/fconv**2, iN, uht)
# Read of the flux in mJy
flux  = rho.at(pos)/kappa.at(pos)
dflux = kappa.at(pos)**-0.5
print("%-15s %8.3f %8.3f %8.3f" % ("lowcorr white", flux, dflux, flux/dflux))
# lowcorr white      7.487    0.711   10.537

# 4. same, but with the other approximation
rho, kappa = analysis.matched_filter_constcorr_smoothivar(map*fconv, beam, ivar/fconv**2, iN, uht)
# Read of the flux in mJy
flux  = rho.at(pos)/kappa.at(pos)
dflux = kappa.at(pos)**-0.5
print("%-15s %8.3f %8.3f %8.3f" % ("smooth white", flux, dflux, flux/dflux))
# smooth white       7.487    0.711   10.537

# 5. 1/f noise power spectrum with l_knee at 2000
iN   =  10**-2/utils.arcmin**2 / (1 + ((uht.l+0.5)/2000)**-3)
rho, kappa = analysis.matched_filter_constcov(map*fconv, beam, iN/fconv**2, uht)
flux  = rho.at(pos)/kappa # kappa just a number in this case
dflux = kappa**-0.5
print("%-15s %8.3f %8.3f %8.3f" % ("constcov 1/f", flux, dflux, flux/dflux))
# constcov 1/f       7.486    0.782    9.571

# 6. 1/f noise and position-dependent depth with lowcorr
ivar  = 10**-2*pixarea/utils.arcmin**2 # Base depth
# spatial modulation with 5 arcmin wavelength horizontal sine wave
ivar *= (1+0.9*np.sin(enmap.posmap(shape, wcs)[1]/(5*utils.arcmin)))
# 1/f spectrum, but dimensionless since ivar handles the units
iN   =  1 / (1 + ((uht.l+0.5)/2000)**-3)
rho, kappa = analysis.matched_filter_constcorr_lowcorr(map*fconv, beam, ivar/fconv**2, iN, uht)
# Read of the flux in mJy
flux  = rho.at(pos)/kappa.at(pos)
dflux = kappa.at(pos)**-0.5
print("%-15s %8.3f %8.3f %8.3f" % ("lowcorr full", flux, dflux, flux/dflux))
# lowcorr full       7.491    0.782    9.578

# 7. same, but with smoothivar
rho, kappa = analysis.matched_filter_constcorr_smoothivar(map*fconv, beam, ivar/fconv**2, iN, uht)
# Read of the flux in mJy
flux  = rho.at(pos)/kappa.at(pos)
dflux = kappa.at(pos)**-0.5
print("%-15s %8.3f %8.3f %8.3f" % ("smooth full", flux, dflux, flux/dflux))
# smooth full        7.483    0.782    9.568

# 8. measuring iN from noise realizations. First build our sims
nsim   = 100
white  = enmap.rand_gauss((nsim,)+shape, wcs)
noise  = uht.harm2map(uht.map2harm(white)*iN**-0.5)*ivar**-0.5
# Measure the 2d noise power spectrum after whitening with ivar,
# and take the mean over all our sims to reduce sample variance.
# Multiplying by the mean pixel size is necessary to get the units right.
iNemp  = 1/(np.mean(np.abs(uht.map2harm(noise*ivar**0.5))**2,0) / noise.pixsize())
# If you measure iN from noise realizations that have very close to zero mean, then
# iNemp may end up with a huge value in the [0,0] (DC) component, which can cause
# issues. In that case you may need to set iNemp[0,0] = 0.
del white, noise
# Use iNemp to matched filter
rho, kappa = analysis.matched_filter_constcorr_lowcorr(map*fconv, beam, ivar/fconv**2, iNemp, uht)
# Read of the flux in mJy
flux  = rho.at(pos)/kappa.at(pos)
dflux = kappa.at(pos)**-0.5
print("%-15s %8.3f %8.3f %8.3f" % ("lowcorr full emp", flux, dflux, flux/dflux))
# lowcorr full emp   7.491    0.778    9.626

# If you don't have a a large number of sims, you can try smoothing N before inverting
# it to from iN, but this can be tricky if N has too much contrast, like a very small
# region with very high values.
"""

import numpy as np, time
from scipy import ndimage
from . import enmap, utils, uharm, wavelets, bunch

# TODO: Figure out the proper way to handle fourier-space
# linear operators. Currently I have a mess of map2harm, harm2map
# map2harm_adjoint, harm2map_adjoint, pixarea and quad_weights.
# It works to the percent level, but it should be better than that.

def matched_filter_constcov(map, B, iN, uht=None, spin=0):
	"""Apply a matched filter to the given map, assuming a constant covariance
	noise model. A constant covariance noise model is one where the pixel-pixel
	covariance is independent of the position in the map, and can be represented
	using a diagonal noise matrix in harmonic space.

	Arguments:
	map:  The [...,ny,nx] enmap to be filtered
	B:    The instrumental beam in the "hprof" format of the provided UHT
	iN:   The inverse noise power in harmonic space, in "hprof" format.
	uht:  The unified harmonic transform (UHT) to use

	Returns rho, kappa, which are [...,ny,nx] enmaps that can be used to construct:
	flux  = rho/kappa
	dflux = kappa**-0.5
	snr   = rho/kappa**0.5

	Here flux is the flux estimate in each pixel. If map is in Jy/sr, then flux
	will be in Jy. dflux is the 1 sigma flux uncertainty. snr is a map of the
	signal-to-noise ratio.
	"""
	# Npix = < n_pix n_pix' > = < Y n_harm n_harm' Y'> = Y Nharm Y'
	# Npix" = Y'" Nharm" Y", Y" = Y'W
	# Npix" = WY Nharm" Y'W. Not the same as YNharm"Y"!
	if uht is None: uht = uharm.UHT(map.shape, map.wcs)
	pixarea = enmap.pixsizemap(map.shape, map.wcs, broadcastable=True)
	rho     = uht.map2harm_adjoint(uht.hmul(B*iN,uht.map2harm(map, spin=spin)),spin=spin)/pixarea
	kappa   = uht.sum_hprof(B**2*iN)/(4*np.pi)
	return rho, kappa

def matched_filter_white(map, B, ivar, uht=None, B2=None, high_acc=False):
	"""Apply a matched filter to the given map, assuming a constant correlation
	noise model inv(N) = ivar, where ivar = 1/pixel_variance.
	This represents pixel-uncorrelated noise.

	Arguments:
	map:  The [...,ny,nx] enmap to be filtered
	B:    The instrumental beam in the "hprof" format of the provided UHT
	ivar: The inverse of the white noise power per pixel, an [...,ny,nx] enmap
	uht:  The unified harmonic transform (UHT) to use
	B2:   The *real-space* square of the beam in "hprof" format. Make sure that the
	      beam is properly normalized to have a unit integral before squaring.
	      Optional. If it is missing, the square will be performed internally.

	Returns rho, kappa, which are [...,ny,nx] enmaps that can be used to construct:
	flux  = rho/kappa
	dflux = kappa**-0.5
	snr   = rho/kappa**0.5

	Here flux is the flux estimate in each pixel. If map is in Jy/sr, then flux
	will be in Jy. dflux is the 1 sigma flux uncertainty. snr is a map of the
	signal-to-noise ratio.
	"""
	# m = YBY"Pa + n, N" = ivar
	# rho   = (YBY"P)'N"m = P'Y"'BY' ivar m = P'WYBY"/W ivar m. Makes sense: converts to intensive units before harmonic
	# kappa = (YBY"P)'N"(YBY"P)
	#       = P'(YBY")'ivar(YBY")P
	# kappa_ii = P_ii²*sum_j (YBY")_ji ivar_jj (YBY")_ji
	#          = P_ii²*sum_j (YBY")_ji² ivar_jj
	#          = P_ii²*sum_j (Y"'BY')_ij² ivar_jj
	P = 1/enmap.pixsizemap(map.shape, map.wcs, broadcastable=True)
	# Square the beam in real space if not provided
	if uht is None: uht = uharm.UHT(map.shape, map.wcs)
	if B2  is None: B2  = uht.hprof_rpow(B, 2)
	# TODO: kappa should have P**2. Figure out why P**1 is what works.
	rho   = P*uht.map2harm_adjoint(uht.hmul(B ,uht.harm2map_adjoint(ivar*map)))
	kappa = P*uht.map2harm_adjoint(uht.hmul(B2,uht.harm2map_adjoint(ivar)))
	return rho, kappa

def matched_filter_constcorr_lowcorr(map, B, ivar, iC, uht=None, B2=None, high_acc=False,
		S=None, iS=None):
	"""Apply a matched filter to the given map, assuming a constant correlation
	noise model inv(N) = ivar**0.5 * iC * ivar**0.5, where ivar = 1/pixel_variance
	and iC = 1/harmonic_power(noise*ivar**0.5). This represents correlated noise
	described by iC that's modulated spatially by ivar.

	Arguments:
	map:  The [...,ny,nx] enmap to be filtered
	B:    The instrumental beam in the "hprof" format of the provided UHT
	ivar: The inverse of the white noise power per pixel, an [...,ny,nx] enmap
	iC:   The inverse power spectrum of the whitened map, map/ivar**0.5, as computed using
	      uht. Note: This will be 1/pixsize for a white noise map, not 1, due to
	      how the fourier space units work.
	uht:  The unified harmonic transform (UHT) to use
	B2:   The *real-space* square of the beam in "hprof" format. Make sure that the
	      beam is properly normalized to have a unit integral before squaring.
	      Optional. If it is missing, the square will be performed internally.
	high_acc: We approximate kappa internally. If high_acc is True, then the exacat
	      kappa will be evaluated for a single pixel, and used to rescale the approximate
	      kappa.

	Returns rho, kappa, which are [...,ny,nx] enmaps that can be used to construct:
	flux  = rho/kappa
	dflux = kappa**-0.5
	snr   = rho/kappa**0.5

	Here flux is the flux estimate in each pixel. If map is in Jy/sr, then flux
	will be in Jy. dflux is the 1 sigma flux uncertainty. snr is a map of the
	signal-to-noise ratio.

	Internally we assume that we can replace iC with its beam²-weighted
	average when computing kappa. This is a good approximation as long as
	iC isn't too steep. Handles rapid variations and holes in ivar well,
	though a hole in the middle of a source can be 20% wrong. That's still
	much better than the smoothivar approximation.
	"""
	# m = YBY"Pa + n
	# rho   = P' Y"'B'Y' V Y'" Ch" Y" V m
	# kappa = P'Y"'B'Y'  V Y'" Ch" Y" V  YBY"P
	# Pretend that we can set Ch" = alpha in kappa
	# kappa approx alpha * P'Y"'B'Y' ivar YBY"P
	# kappa_ii = alpha*P_ii²*sum_j (YBY")_ji² ivar_jj
	#  = alpha * convolve(br**2, ivar)_ii
	if uht is None: uht = uharm.UHT(map.shape, map.wcs)
	pixarea = enmap.pixsizemap(map.shape, map.wcs, broadcastable=True)
	V = ivar**0.5
	W = uht.quad_weights()
	# Square the beam in real space if not provided
	if B2 is None: B2 = uht.hprof_rpow(B, 2)
	# Optional skew-like transformation of covariance. Only applies to rho
	# under the lowcorr approximation
	if S  is None: S  = lambda x: x
	if iS is None: iS = lambda x: x
	# Find a white approximation for iC. A B²-weighted average is accurate to
	# about 1% for lknee = 4000, worsening to about 3% by lknee = 7000. Probably
	# good enough.
	iC_white = uht.sum_hprof(B**2*iC)/uht.sum_hprof(B**2)

	rho   = uht.harm2map(uht.hmul(B,uht.harm2map_adjoint(V*iS(uht.map2harm_adjoint(uht.hmul(iC, uht.map2harm(S(V*map))))))))/pixarea
	kappa = uht.map2harm_adjoint(uht.hmul(B2,uht.harm2map_adjoint(ivar*W*iC_white[...,None,None])))/pixarea**2

	if high_acc:
		# Optionally find a correction factor by evaluating the exact kappa in a single pixel
		pix  = tuple(np.array(map.shape[-2:])//2)
		u    = map*0; u[...,pix[0],pix[1]] = 1
		kappa_ii = (uht.harm2map(uht.hmul(B,uht.harm2map_adjoint(V*uht.map2harm_adjoint(uht.hmul(iC,uht.map2harm(V*uht.harm2map(uht.hmul(B,uht.map2harm(u/pixarea)))))))))/pixarea)[...,pix[0],pix[1]]
		alpha  = kappa[...,pix[0],pix[1]]/kappa_ii
		kappa /= alpha[...,None,None]

	return rho, kappa

def matched_filter_constcorr_smoothivar(map, B, ivar, iC, uht=None):
	"""Apply a matched filter to the given map, assuming a constant correlation
	noise model inv(N) = ivar**0.5 * iC * ivar**0.5, where ivar = 1/pixel_variance
	and iC = 1/harmonic_power(noise*ivar**0.5). This represents correlated noise
	described by iC that's modulated spatially by ivar.

	Arguments:
	map:  The [...,ny,nx] enmap to be filtered
	B:    The instrumental beam in the "hprof" format of the provided UHT
	ivar: The inverse of the white noise power per pixel, an [...,ny,nx] enmap
	iC:   The inverse power spectrum of the whitened map, map/ivar**0.5, as computed using
	      uht. Note: This will be 1/pixsize for a white noise map, not 1, due to
	      how the fourier space units work.
	uht:  The unified harmonic transform (UHT) to use

	Returns rho, kappa, which are [...,ny,nx] enmaps that can be used to construct:
	flux  = rho/kappa
	dflux = kappa**-0.5
	snr   = rho/kappa**0.5

	Here flux is the flux estimate in each pixel. If map is in Jy/sr, then flux
	will be in Jy. dflux is the 1 sigma flux uncertainty. snr is a map of the
	signal-to-noise ratio.

	Internally we assume that the beam can be commuted past V, such that
	B'N" approx V"B'C"V". This gives us an easy analytic expression for kappa,
	but implicitly assumes that the hitcount doesn't change very rapidly
	from pixel to pixel. It breaks down completely if there is a hole at the
	peak of a source.
	"""
	# See the constcov function for a bit more math details.
	# We assume that we can commute B past V, allowing us to compute kappa directly
	if uht is None: uht = uharm.UHT(map.shape, map.wcs)
	V    = ivar**0.5
	P    = 1/enmap.pixsizemap(map.shape, map.wcs, broadcastable=True)
	rho  = P*V*uht.map2harm_adjoint(uht.hmul(B*iC,uht.harm2map_adjoint(V*map)))
	# kappa = P'VY"'B C" B Y"VP = (P'V = R)(Y"'B sqrt(CH") = A)A'R' = RAA'R'
	# kappa_ii = R_ii² sum_l A_il
	kappa = ivar * (uht.sum_hprof(B**2*iC)/(4*np.pi))[...,None,None]*P
	return rho, kappa

def matched_filter_constcorr_dual(map, B, ivar, iC, uht=None, S=None, iS=None):
	"""Apply a matched filter to the given map, assuming a dual constant correlation
	noise model inv(N) = iC**0.5 * ivar * iC**0.5, where ivar = 1/pixel_variance
	and iC = 1/harmonic_power(noise*ivar**0.5). This represents correlated noise
	described by iC that's modulated spatially by ivar, but does so differently
	from the standard constcorr method. This may be worse at handling sharp edges,
	but should avoid the problem where hitcount-variations induce large-scale signal
	in the filterd maps.

	Arguments:
	map:  The [...,ny,nx] enmap to be filtered
	B:    The instrumental beam in the "hprof" format of the provided UHT
	ivar: The inverse of the white noise power per pixel, an [...,ny,nx] enmap
	iC:   The inverse power spectrum of the whitened map, map/ivar**0.5, as computed using
	      uht. Note: This will be 1/pixsize for a white noise map, not 1, due to
	      how the fourier space units work.
	uht:  The unified harmonic transform (UHT) to use

	Returns rho, kappa, which are [...,ny,nx] enmaps that can be used to construct:
	flux  = rho/kappa
	dflux = kappa**-0.5
	snr   = rho/kappa**0.5

	Here flux is the flux estimate in each pixel. If map is in Jy/sr, then flux
	will be in Jy. dflux is the 1 sigma flux uncertainty. snr is a map of the
	signal-to-noise ratio.
	"""
	if uht is None: uht = uharm.UHT(map.shape, map.wcs)
	pixarea = enmap.pixsizemap(map.shape, map.wcs, broadcastable=True)
	W = uht.quad_weights()
	# Square root of iC. This does not handle [ncomp,ncomp,...]-type
	# iCs, which would need something like eigpow
	hC = iC**0.5
	# Real-space square of B*iC**0.5
	BC2 = uht.hprof_rpow(B*hC, 2)
	# Optional skew-like transformation of covariance. Only applies to rho
	# since we can't handle it in kappa, but it should be a pretty good approximation
	if S  is None: S  = lambda x: x
	if iS is None: iS = lambda x: x
	# fc = fC**0.5 r
	# c  = harm2map fc = harm2map fC**0.5 r
	# C  = <cc'> = harm2map fC**0.5 I fC**0.5 harm2map' = harm2map fC harm2map'
	# C**2 = harm2map fC harm2map' harm2map fC harm2map' != map2harm fC**2 map2harm'
	# I tried deriving the right mix of normal vs. adjoint operations here, but what
	# gives the smallest bias in practice is the straightforward below.
	# curved doesn't really care, while flat goes from 30% bias in 50-60 el patch
	# to 6% bias.
	rho   = uht.harm2map(uht.hmul(B,uht.map2harm(iS(uht.harm2map(uht.hmul(hC,uht.map2harm(S(ivar*iS(uht.harm2map(uht.hmul(hC,uht.map2harm(S(map)))))))))))))*pixarea**-1
	kappa = uht.map2harm_adjoint(uht.hmul(BC2,uht.harm2map_adjoint(ivar*W)))/pixarea**2
	return rho, kappa


# These functions and classes represent a modular approach to
# object detection in maps. At the lowest level are Nmat objects,
# which provide matched filtering through a unified interface.
# Then there are Finder objects that are typically built using an
# Nmat object and returns a catalog given a map.
# There are also Measurer objects that are similar to Finders,
# but take a catalog of known positions as input.
# Finally, there are Modellers that go from a catalog to a map.
#
# The nice thing with this approach is that it's easy to compose
# these together to make more advanced things, like iterative
# finders etc.
#
# A limitation of the implementation below is that they're
# flat-sky only, unlike the UHT-based functions above. For me,
# that hasn't turned out to be a limitation, since I have to do
# things in tiles for other reasons anyway (e.g. memory).


# Overall abstract interfaces

class Nmat:
	def matched_filter(self, map, cache=None): raise NotImplementedError
	def simulate(self): raise NotImplementedError

class Finder:
	def __call__(self, map): raise NotImplementedError

class Measurer:
	def __call__(self, map, cat): raise NotImplementedError

class Modeller:
	def __call__(self, cat): raise NotImplementedError
	def amplitudes(self, cat): raise NotImplementedError

# Nmat implementations:

class NmatConstcov(Nmat):
	def __init__(self, iN, apod):
		"""Initialize a Constcov noise model from an inverse noise power spectrum
		enmap. iN must have shape [n,n,ny,nx]. For a simple scalar filter just
		insert scalar dimensions with None before constructing, e.g.
		iN[None,None]."""
		self.iN      = iN
		self.apod    = apod
		assert self.iN.ndim == 4, "iN   must be an enmap with 4 dims"
		self.pixsize  = enmap.pixsize(self.iN.shape, self.iN.wcs)
		self.pixratio = enmap.pixsizemap(self.iN.shape, self.iN.wcs, broadcastable=True)/self.pixsize
		self.fsky    = enmap.area(self.iN.shape, self.iN.wcs)/(4*np.pi)
	def matched_filter(self, map, beam, cache=None):
		"""Apply a matched filter to the given map [n,ny,nx], which must agree in shape
		with the beam transform beam [n,ny,nx]. Returns rho[n,ny,nx], kappa[n,n,ny,nx].
		From these the best-fit fluxes in pixel y,x can be recovered as
		np.linalg.solve(kappa[:,:,y,x],rho[:,y,x]), and the combined flux as
		rho_tot = np.sum(rho,0); kappa_tot = np.sum(kappa,(0,1)); flux_tot = rho_tot[y,x]/kappa_tot[y,x]"""
		assert map .ndim  == 3, "Map must be an enmap with 3 dims"
		assert beam.ndim  == 3, "Beam must be an enmap with 3 dims"
		assert map .shape == beam.shape, "Map and beam shape must agree"
		# Flat sky corrections. Don't work that well here. Without them
		# we have an up to 2% error with a 10° tall patch. With them this
		# is reduced to 1%. I don't understand why they don't work better here.
		flatcorr_rho, flatcorr_kappa = get_flat_sky_correction(self.pixratio)
		rho = utils.cache_get(cache, "rho_pre", lambda: enmap.map_mul(self.iN,enmap.map2harm(map*self.apod, spin=0, normalize="phys"))/self.pixsize)
		rho = enmap.map2harm_adjoint(beam*rho, spin=0, normalize="phys")*flatcorr_rho
		kappa0 = np.sum(beam[:,None]*self.iN[:,:]*beam[None,:],(-2,-1))/(4*np.pi*self.fsky)
		kappa  = np.empty_like(rho, shape=kappa0.shape+rho.shape[-2:])
		kappa[:] = kappa0[:,:,None,None]*flatcorr_kappa
		# Done! What we return will always be [n,ny,nx], [n,n,ny,nx]
		return rho, kappa
	def simulate(self):
		hN = safe_pow(self.iN, -0.5)
		r  = enmap.rand_gauss_harm(self.iN.shape[1:], self.iN.wcs).astype(utils.complex_dtype(self.iN.dtype))
		sim= enmap.ifft(enmap.map_mul(hN, r), normalize="phys").real
		return sim

class NmatConstcorr(Nmat):
	def __init__(self, iN, ivar):
		"""Initialize a Constcov noise model from an inverse noise power spectrum
		enmap. iN must have shape [n,n,ny,nx]. For a simple scalar filter just
		insert scalar dimensions with None before constructing, e.g.
		iN[None,None]."""
		self.iN      = iN
		self.ivar    = ivar
		assert self.iN  .ndim == 4, "iN   must be an enmap with 4 dims"
		assert self.ivar.ndim == 3, "ivar must be an enmap with 3 dims"
		self.pixsize  = enmap.pixsize(self.iN.shape, self.iN.wcs)
		self.pixratio = enmap.pixsizemap(self.iN.shape, self.iN.wcs, broadcastable=True)/self.pixsize
	def matched_filter(self, map, beam, beam2=None, cache=None):
		"""Apply a matched filter to the given map [n,ny,nx], which must agree in shape
		with the beam transform beam [n,ny,nx]. Returns rho[n,ny,nx], kappa[n,n,ny,nx].
		From these the best-fit fluxes in pixel y,x can be recovered as
		np.linalg.solve(kappa[:,:,y,x],rho[:,y,x]), and the combined flux as
		rho_tot = np.sum(rho,0); kappa_tot = np.sum(kappa,(0,1)); flux_tot = rho_tot[y,x]/kappa_tot[y,x]"""
		assert map.ndim == 3, "Map must be an enmap with 3 dims"
		assert beam.ndim == 3, "Beam must be an enmap with 3 dims"
		assert map.shape == beam.shape, "Map and beam shape must agree"
		V    = self.ivar**0.5
		# Square the beam in real space if not provided
		if beam2 is None: beam2 = rpow(beam, 2)
		# Find a white approximation for iN. Is doing this element-wise correct?
		iN_white = np.sum(beam[:,None]*self.iN*beam[None,:],(-2,-1))/np.sum(beam[:,None]*beam[None,:],(-2,-1))
		# Flat-sky correction factors. For a 10° tall patch we get up to 2% errors in flux without
		# them. This is reduced to 1% with them. Not sure why it doesn't do better - is the
		# gaussian approximation too limiting? Or do I have a bug?
		flatcorr_rho, flatcorr_kappa = get_flat_sky_correction(self.pixratio)
		# Numerator
		rho   = utils.cache_get(cache, "rho_pre", lambda: enmap.map2harm(flatcorr_rho*V*enmap.harm2map(enmap.map_mul(self.iN, enmap.map2harm(V*map, spin=0, normalize="phys")), spin=0, normalize="phys"), spin=0, normalize="phys")/self.pixsize)
		rho   = enmap.harm2map(beam*rho, spin=0, normalize="phys")
		# Denominator
		kappa = enmap.harm2map(enmap.map_mul(beam2,enmap.map2harm(self.ivar+0j, spin=0, normalize="phys")), spin=0, normalize="phys")/self.pixsize * flatcorr_kappa
		kappa = np.maximum(kappa,0)**0.5
		kappa = kappa[:,None]*iN_white[:,:,None,None]*kappa[None,:]
		# Done! What we return will always be [n,ny,nx], [n,n,ny,nx]
		return rho, kappa
	def simulate(self):
		hN = safe_pow(self.iN, -0.5)
		r  = enmap.rand_gauss_harm(self.iN.shape[1:], self.iN.wcs).astype(utils.complex_dtype(self.iN.dtype))
		sim= enmap.ifft(enmap.map_mul(hN, r)).real
		mask = self.ivar != 0
		sim[mask] *= self.ivar[mask]**-0.5
		return sim

class NmatWavelet(Nmat):
	def __init__(self, wt, wiN):
		"""Initialize a Wavelet noise model from a WaveletTransform object and
		a wavelet map object representing the inverse noise variance map pwer wavelet scale."""
		self.wt   = wt
		self.wiN  = wiN
	def matched_filter(self, map, beam, cache=None):
		# We get 2% flat-sky errors with a 10° tall patch without corrections, and 1% with the corrections.
		pixsize  = enmap.pixsize(map.shape, map.wcs)
		pixratio = enmap.pixsizemap(map.shape, map.wcs, broadcastable=True)/pixsize
		flatcorr_rho, flatcorr_kappa = get_flat_sky_correction(pixratio)
		# Get rho
		rho = utils.cache_get(cache, "rho_pre", lambda: enmap.map2harm(self.wt.wave2map(multimap.map_mul(self.wiN, self.wt.map2wave(map))), spin=0, normalize="phys")/pixsize)
		rho = enmap.harm2map(beam*rho, spin=0, normalize="phys")*flatcorr_rho
		# Then get kappa
		fkappa = enmap.zeros(self.wiN.pre + map.shape[-2:], map.wcs, utils.complex_dtype(map.dtype))
		for i in range(self.wt.nlevel):
			sub_Q  = self.wt.filters[i]*enmap.resample_fft(beam, self.wt.geometries[i][0], norm=None, corner=True)
			# Is it right to do this component-wise?
			sub_Q2 = rop(sub_Q, op=lambda a: a[:,None]*a[None,:])
			fsmall = sub_Q2*enmap.fft(self.wiN.maps[i], normalize=False)/self.wiN.npixs[i]
			enmap.resample_fft(fsmall, map.shape, fomap=fkappa, norm=None, corner=True, op=np.add)
		kappa = enmap.ifft(fkappa, normalize=False).real/pixsize*flatcorr_kappa
		return rho, kappa


# Finder implementations

class FinderSimple(Finder):
	def __init__(self, nmat, beam, scaling=1, save_snr=False):
		"""Initialize a simple object finder. It looks for a single class
		of object in maps with shape [nfreq,ny,nx]. It looks for objects
		with shape given by the 2d fourier profile "beam". For point
		soruces this will be an actual beam transform. For extended objects
		it will be the product of the beam transform and their shape's fourier coefficients.

		The argument "scaling" specifies how the signal is expected to change
		between frequencies, and should be either a scalar or an array of shape [nfreq].

		"nmat" should be an Nmat subclass that describes the noise properties of
		the maps to be analysed.

		This class does not do anything fancy like iterative finding etc.
		This means that it won't handle e.g. weak objects close to strong objects
		well. It also looks for just a single type of object. Use FinderIterative
		or FinderMulti (or both) to get around this."""
		self.beam   = beam
		self.nmat   = nmat
		self.scaling= scaling
		self.order  = 3
		self.grow   = 0.75*utils.arcmin
		self.grow0  = 20*utils.arcmin
		self.save_snr = save_snr
		self.snr    = None
	def __call__(self, map, snmin=5, snrel=None, penalty=None):
		assert map.ndim == 3, "Map must be an enmap with 3 dims"
		ncomp = len(map)
		dtype = [("ra","d"),("dec","d"),("snr","d"),("flux_tot","d"),("dflux_tot","d"),
				("flux","d",(ncomp,)),("dflux","d",(ncomp,))]
		if penalty is None: penalty = 1
		# Apply the matched filter
		rho, kappa = self.nmat.matched_filter(map, self.beam)
		kappa     = sanitize_kappa(kappa)
		scaling   = np.zeros(len(rho),rho.dtype)+self.scaling
		rho_tot   = np.sum(rho*scaling[:,None,None],0)
		# Build the total detection significance and find peaks
		kappa_tot = np.sum(scaling[:,None,None,None]*kappa*scaling[None,:,None,None],(0,1))
		snr_tot   = rho_tot/kappa_tot**0.5
		# Find the effective S/N threshold, taking into account any position-dependent
		# penalty and the (penalized) maximum value in the map
		if snrel   is not None: snmin = max(snmin, np.max(snr_tot/penalty)*snrel)
		snlim = snmin*penalty
		# Detect objects, and grow them a bit
		labels, nlabel = ndimage.label(snr_tot >= snlim)
		cat            = np.zeros(nlabel, dtype).view(np.recarray)
		if nlabel == 0: return bunch.Bunch(cat=cat, snmin=snmin, snr=snr_tot, snlim=snlim)
		labels = enmap.samewcs(labels, map)
		dists, labels = labels.labeled_distance_transform(rmax=self.grow0)
		labels *= dists <= self.grow
		allofthem = np.arange(1,nlabel+1)
		if len(cat) == 0: return bunch.Bunch(cat=cat, snmin=snmin, snr=snr_tot, snlim=snlim)
		# Find the position and snr of each object
		pixs    = np.array(ndimage.center_of_mass(snr_tot**2, labels, allofthem)).T
		cat.ra, cat.dec = map.pix2sky(pixs)[::-1]
		cat.snr = ndimage.maximum(snr_tot, labels, allofthem)
		del labels
		# Interpolating before solving is faster, but inaccurate. So we do the slow thing.
		# First get the total flux and its uncertainty.
		flux_tot, dflux_tot = solve_mapsys(kappa_tot, rho_tot)
		del rho_tot, kappa_tot
		cat.flux_tot        = flux_tot .at(pixs, unit="pix", order=self.order)
		cat.dflux_tot       = dflux_tot.at(pixs, unit="pix", order=0)
		del flux_tot, dflux_tot
		# Then get the per-freq versions
		flux, dflux = solve_mapsys(kappa, rho)
		del rho, kappa
		cat.flux    = flux .at(pixs, unit="pix", order=self.order).T
		cat.dflux   = dflux.at(pixs, unit="pix", order=0).T
		del flux, dflux
		# Hack
		if self.save_snr and self.snr is None: self.snr = snr_tot
		# Sort by SNR and return
		cat = cat[np.argsort(cat.snr)[::-1]]
		return bunch.Bunch(cat=cat, snmin=snmin, snr=snr_tot, snlim=snlim)

moo = 0
class FinderMulti(Finder):
	def __init__(self, nmat, beams, scalings, save_snr=False):
		"""Initialize FinderMulti. It's similar to FinderSimple, except
		that it looks for multiple different types of object at the same
		time. "beams" is a list of 2d fourier representations of shapes
		to look for, while scalings is a list frequency scalings with
		the same length."""
		self.nmat     = nmat
		self.beams    = beams
		if scalings is None:
			scalings = np.ones(len(beams))
		self.scalings = scalings
		self.order    = 3
		self.r        = 2*utils.arcmin
		self.save_snr = save_snr
		self.snr      = None
	def __call__(self, map, snmin=5, snrel=None, penalty=None):
		global moo; moo += 1
		#enmap.write_map("map_%02d.fits" % moo, map)
		assert map.ndim == 3, "Map must be an enmap with 3 dims"
		ncase = len(self.beams)
		ncomp = len(map)
		dtype = [("ra","d"),("dec","d"),("snr","d"),("flux_tot","d"),("dflux_tot","d"),
				("flux","d",(ncomp,)),("dflux","d",(ncomp,)),("case","i")]
		if penalty is None: penalty = 1
		# Apply the matched filter for each profile and keep track of the best beam
		# in each pixel.
		cache = {}
		snr_tot, rho, kappa, rho_tot, kappa_tot, case = None, None, None, None, None, None
		for ca, (beam, scaling) in enumerate(zip(self.beams, self.scalings)):
			def f():
				rho, kappa = self.nmat.matched_filter(map, beam, cache=cache)
				kappa      = sanitize_kappa(kappa)
				return rho, kappa
			my_rho, my_kappa = utils.cache_get(cache, "beam:"+str(id(beam)), f)
			my_rho_tot    = np.sum(my_rho*scaling[:,None,None],0)
			my_kappa_tot  = np.sum(scaling[:,None,None,None]*my_kappa*scaling[None,:,None,None],(0,1))
			my_snr_tot    = my_rho_tot/my_kappa_tot**0.5
			#enmap.write_map("snr_%02d_case_%02d.fits" % (moo, ca), my_snr_tot)
			if snr_tot is None:
				cases = enmap.full(my_snr_tot.shape, my_snr_tot.wcs, ca, np.int8)
				snr_tot, rho, kappa, rho_tot, kappa_tot= my_snr_tot, my_rho, my_kappa, my_rho_tot, my_kappa_tot
			else:
				mask = my_snr_tot > snr_tot
				cases      = enmap.samewcs(np.where(mask, ca,           cases),      map)
				snr_tot    = enmap.samewcs(np.where(mask, my_snr_tot,   snr_tot),   map)
				rho        = enmap.samewcs(np.where(mask, my_rho,       rho),       map)
				kappa      = enmap.samewcs(np.where(mask, my_kappa,     kappa),     map)
				rho_tot    = enmap.samewcs(np.where(mask, my_rho_tot,   rho_tot),   map)
				kappa_tot  = enmap.samewcs(np.where(mask, my_kappa_tot, kappa_tot), map)
			del my_rho, my_kappa, my_rho_tot, my_kappa_tot, my_snr_tot
		del cache
		#enmap.write_map("snr_%02d.fits"  % (moo), snr_tot)
		#enmap.write_map("case_%02d.fits" % (moo), cases)
		# Hack
		if self.save_snr and self.snr is None: self.snr = snr_tot
		# Find the effective S/N threshold, taking into account any position-dependent
		# penalty and the (penalized) maximum value in the map
		if snrel   is not None: snmin = max(snmin, np.max(snr_tot/penalty)*snrel)
		snlim = snmin*penalty
		labels, nlabel = ndimage.label(snr_tot >= snlim)
		allofthem      = np.arange(1,nlabel+1)
		cat            = np.zeros(nlabel, dtype).view(np.recarray)
		if nlabel == 0: return bunch.Bunch(cat=cat, snmin=snmin, snr=snr_tot, snlim=snlim)
		pixs0   = np.array(ndimage.maximum_position(snr_tot, labels, allofthem)).T
		# Make the labels circles with a constant radius. This ensures that the
		# center-of-mass is calculated over a consistent area.
		labels = enmap.samewcs(labels, map)
		labels = make_circle_labels(map.shape, map.wcs, pixs0, r=self.r)
		#enmap.write_map("labels_%02d.fits" % (moo), labels)
		if len(cat) == 0: return bunch.Bunch(cat=cat, snmin=snmin, snr=snr_tot, snlim=snlim)
		# Find the position and snr of each object. This is a bit questionable since
		# only the central pixel is guaranteed to belong to relevant case. We therefore
		# check that the center-of-mass location has the same case, and if it doesn't,
		# we fall back on just using the pixel center. This is an acceptable solution,
		# but not a good one. It means that fluxes and positions will be inaccurate for
		# the cases where we do end up taking the fallback. Hopefully those should be
		# pretty weak cases anyway. I think a better solution would require the individual
		# case information, e.g. via a two-pass approach or by saving per-case flux, dflux,
		# snr etc. maps.
		pixs    = np.array(ndimage.center_of_mass(snr_tot**2, labels, allofthem)).T
		cat.snr = ndimage.maximum(snr_tot, labels, allofthem)
		# Interpolating before solving is faster, but inaccurate. So we do the slow thing.
		flux_tot, dflux_tot = solve_mapsys(kappa_tot, rho_tot)
		case0    = cases.at(pixs0, unit="pix", order=0)
		case_com = cases.at(pixs,  unit="pix", order=0)
		flux0    = flux_tot.at(pixs0, unit="pix", order=0)
		flux_com = flux_tot.at(pixs,  unit="pix", order=self.order)
		unsafe   = (case_com != case0) | (np.abs((flux_com-flux0))/np.maximum(np.abs(flux_com),np.abs(flux0)) > 0.2)
		# Build the total part of the catalog
		cat.ra, cat.dec = map.pix2sky(np.where(unsafe, pixs0, pixs))[::-1]
		cat.case      = np.where(unsafe, case0, case_com)
		cat.flux_tot  = np.where(unsafe, flux0, flux_com)
		cat.dflux_tot = dflux_tot.at(np.where(unsafe, pixs0, pixs), unit="pix", order=0)
		del flux_tot, dflux_tot
		# Then get the per-freq versions
		flux, dflux = solve_mapsys(kappa, rho)
		del rho, kappa
		cat.flux    = np.where(unsafe, flux.at(pixs0, unit="pix", order=0), flux.at(pixs, unit="pix", order=self.order)).T
		cat.dflux   = dflux.at(np.where(unsafe, pixs0, pixs), unit="pix", order=0).T
		del flux, dflux
		# Sort by SNR and return
		cat = cat[np.argsort(cat.snr)[::-1]]
		return bunch.Bunch(cat=cat, snmin=snmin, snr=snr_tot, snlim=snlim)

moo = 0
class FinderMultiSafe(Finder):
	def __init__(self, nmat, beams, scalings, save_snr=False):
		"""Initialize FinderMultiSafe. It's similar to FinderSimple, except
		that it looks for multiple different types of object at the same
		time. "beams" is a list of 2d fourier representations of shapes
		to look for, while scalings is a list frequency scalings with
		the same length."""
		self.nmat     = nmat
		self.beams    = beams
		if scalings is None:
			scalings = np.ones(len(beams))
		self.scalings = scalings
		self.order    = 3
		self.rs       = np.array([get_central_radius(beam, lknee=2000) for beam in self.beams])
		print(self.rs/utils.arcmin)
		self.save_snr = save_snr
		self.snr      = None
	def __call__(self, map, snmin=5, snrel=None, penalty=None):
		global moo; moo += 1
		#enmap.write_map("map_%02d.fits" % moo, map)
		assert map.ndim == 3, "Map must be an enmap with 3 dims"
		ncase = len(self.beams)
		ncomp = len(map)
		dtype = [("ra","d"),("dec","d"),("snr","d"),("flux_tot","d"),("dflux_tot","d"),
				("flux","d",(ncomp,)),("dflux","d",(ncomp,)),("case","i")]
		if penalty is None: penalty = 1
		# Apply the matched filter for each profile and keep track of the best beam
		# in each pixel.
		cache = {}
		snr_tot, cases = None, None
		snrs, fluxs_tot, dfluxs_tot, fluxs, dfluxs = [], [], [], [], []
		for ca, (beam, scaling) in enumerate(zip(self.beams, self.scalings)):
			def f():
				rho, kappa = self.nmat.matched_filter(map, beam, cache=cache)
				kappa      = sanitize_kappa(kappa)
				return rho, kappa
			my_rho, my_kappa = utils.cache_get(cache, "beam:"+str(id(beam)), f)
			my_rho_tot    = np.sum(my_rho*scaling[:,None,None],0)
			my_kappa_tot  = np.sum(scaling[:,None,None,None]*my_kappa*scaling[None,:,None,None],(0,1))
			my_snr        = my_rho_tot/my_kappa_tot**0.5
			my_flux,     my_dflux     = solve_mapsys(my_kappa, my_rho)
			my_flux_tot, my_dflux_tot = solve_mapsys(my_kappa_tot, my_rho_tot)
			#enmap.write_map("snr_%02d_case_%02d.fits" % (moo, ca), my_snr)
			if snr_tot is None:
				cases = enmap.full(my_snr.shape, my_snr.wcs, ca, np.int8)
				snr_tot = my_snr
			else:
				mask       = my_snr > snr_tot
				cases      = enmap.samewcs(np.where(mask, ca,     cases),   map)
				snr_tot    = enmap.samewcs(np.where(mask, my_snr, snr_tot), map)
			fluxs_tot .append(my_flux_tot)
			dfluxs_tot.append(my_dflux_tot)
			fluxs .append(my_flux)
			dfluxs.append(my_dflux)
			snrs.append(my_snr)
			del my_rho, my_kappa, my_flux, my_dflux, my_flux_tot, my_dflux_tot, my_snr
		del cache
		#enmap.write_map("snr_%02d.fits"  % (moo), snr_tot)
		#enmap.write_map("case_%02d.fits" % (moo), cases)
		# Hack
		if self.save_snr and self.snr is None: self.snr = snr_tot
		# Find the effective S/N threshold, taking into account any position-dependent
		# penalty and the (penalized) maximum value in the map
		if snrel   is not None: snmin = max(snmin, np.max(snr_tot/penalty)*snrel)
		snlim = snmin*penalty
		labels, nlabel = ndimage.label(snr_tot >= snlim)
		labels         = enmap.samewcs(labels, map)
		allofthem      = np.arange(1,nlabel+1)
		cat            = np.zeros(nlabel, dtype).view(np.recarray)
		if nlabel == 0: return bunch.Bunch(cat=cat, snmin=snmin, snr=snr_tot, snlim=snlim)
		# Find the peak pixel for each label, and use them to reshape each label into
		# a constant-radius circle. TODO: Consider making this radius case-dependent,
		# with a radius given by the filter kernel width.
		pixs0   = np.array(ndimage.maximum_position(snr_tot, labels, allofthem)).T
		#enmap.write_map("labels_%02d.fits" % (moo), labels)
		# Find which case that corresponds to
		cat.case= cases[pixs0[0],pixs0[1]]
		cat.snr = snr_tot[pixs0[0],pixs0[1]]
		# We will now loop over cases and process the objects for each case
		# separately. This avoids the issue of averaging over pixels that correspond
		# to multiple different cases that the plain FinderMulti has.
		for ca in range(ncase):
			someofthem = allofthem[cat.case==ca]
			inds       = someofthem-1
			if len(inds) == 0: continue
			my_labels  = make_circle_labels(map.shape, map.wcs, pixs0[:,inds], inds=someofthem, r=self.rs[ca])
			pixs = np.array(ndimage.center_of_mass(snrs[ca]**2, my_labels, someofthem)).T
			cat.ra[inds], cat.dec[inds] = map.pix2sky(pixs)[::-1]
			cat. flux_tot[inds] =  fluxs_tot[ca].at(pixs, unit="pix", order=self.order)
			cat.dflux_tot[inds] = dfluxs_tot[ca].at(pixs, unit="pix", order=0)
			cat. flux[inds]     =  fluxs[ca].at(pixs, unit="pix", order=self.order).T
			cat.dflux[inds]     = dfluxs[ca].at(pixs, unit="pix", order=0).T
		# Sort by SNR and return
		cat = cat[np.argsort(cat.snr)[::-1]]
		return bunch.Bunch(cat=cat, snmin=snmin, snr=snr_tot, snlim=snlim)

class FinderIterative(Finder):
	def __init__(self, finder, modeller, maxiter=20, sntol=0.50,
			grid_max=5, grid_res=0.1*utils.degree, grid_dominance=4):
		"""Initialize an iterative object finder. This handles weaker
		objects near stronger ones by iteratively finding the strongest
		objects, subtracting them, finding weaker objects, and so on.

		Arguments:
		* finder: The Finder object to use in each step of the iterative
		  analysis, e.g. a FinderSimple or FinderMulti instance.
		* modeller: The Modeller object to use in the subtraction step.
		  Must be consistent with the catalogue returned by the finder.
		  For example, a FinderMulti must be used with a ModellerMulti.
		
		The other arguments are optional:
		* maxiter: The maximum number of iterations to run.
		* sntol:   The factor by which the detection limit decreases by
		  in each iteration.
		* grid_max: The maximum number of objects to allow in a small area.
		* grid_res: The side-length of that small area.
		* grid_dominance: Keep objects in an area that exceeded grid_max
		  if they are contaminated by less than 1/grid_dominance. Put
		  another way, the objects to keep should be at least grid_dominance
		  times higher than their contamination by other objects."""
		self.finder   = finder
		self.modeller = modeller
		self.maxiter  = maxiter
		self.sntol    = sntol
		self.grid_max = grid_max
		self.grid_res = grid_res
		self.grid_dominance = grid_dominance
	def __call__(self, map, snmin=5, snrel=None, verbose=False):
		cat      = []
		own_vals = []
		model = map*0
		# Build a grid that's used to penalize areas with too many detections
		box  = utils.widen_box(map.corners(),2*self.grid_res,relative=False)
		cgeo = enmap.geometry(box, res=self.grid_res)
		hits = enmap.zeros(*cgeo, np.int32)
		snmaxs = enmap.zeros(*cgeo, map.dtype)
		snbad  = enmap.zeros(*cgeo, bool)
		for i in range(self.maxiter):
			t1    = time.time()
			sntol = self.sntol
			bad_mask  = hits > self.grid_max
			bad_mask |= snbad
			if np.any(bad_mask):
				bad_mask[:] = ndimage.distance_transform_edt(1-bad_mask)<=1.5
			bad_mask  = bad_mask.project(*map.geometry, order=0)
			res = self.finder(map-model, snmin=snmin, snrel=sntol, penalty=1+bad_mask*1e6)
			t2  = time.time()
			if verbose:
				print("it %d snmin %5.1f nsrc %6d tot %6d time %6.2f" % (i+1, res.snmin, len(res.cat), len(res.cat)+sum([len(c) for c in cat]), t2-t1))
			if len(res.cat) == 0: break
			# Update our grid
			pix   = hits.sky2pix([res.cat.dec,res.cat.ra])
			ipix  = utils.nint(pix)
			hits += utils.bin_multi(ipix, hits.shape)
			# Measure the total snr in each cell. If this is bigger in one
			# step than it was in the previous step, then we have a model mismatch
			# blow-up. This is rare and I haven't gotten to the bottom of it, but
			# it's best to just prevent like this. Allow some tolerance to allow for
			# sources with similar amplitude very close to each other, where the negative
			# ring around one source would make the other one seem weaker at first.
			blowup_tol = 1.5
			snr_bin = utils.bin_multi(utils.nint(pix), hits.shape, res.cat.snr)
			snbad  |= (snr_bin > snmaxs*blowup_tol) & (snmaxs > 0)
			snmaxs  = np.maximum(snmaxs, snr_bin)
			# Disqualify these entries
			bad     = snbad[ipix[0],ipix[1]]
			res.cat = res.cat[~bad]
			if len(res.cat) == 0: continue
			cat.append(res.cat)
			# update total model
			model += self.modeller(res.cat)
			if res.snmin <= snmin: break
		# The final result will have an extra contamination column
		if len(cat) == 0: cat = res.cat
		else:             cat = np.concatenate(cat, 0).view(np.recarray)
		contam = np.zeros(len(cat), [("contam", "d", (len(map),))]).view(np.recarray)
		if len(cat) > 0:
			# Measure the contamination. This is defined as others' contribution
			# to our peak location relative to our own.
			own_vals = self.modeller.amplitudes(cat)
			tot_vals = model.at([cat.dec,cat.ra]).T
			contam.contam = (tot_vals-own_vals)/np.abs(own_vals)
		cat = merge_arrays([cat, contam]).view(np.recarray)
		# Disqualify sources in overly contaminated regions, unless they dominate alone.
		# Maybe this could be moved into a separate class, but single-pass finders will
		# never get very dense detectors, and the grid is needed here anyway to avoid
		# the finding being too slow.
		if self.grid_dominance and len(cat) > 0:
			# We're bad if the cell as too many hits...
			pix   = utils.nint(hits.sky2pix([cat.dec,cat.ra]))
			nhit  = hits.at(pix, unit="pix", order=0)
			bad   = nhit > self.grid_max
			# ...unless we dominate
			freq_snr    = np.abs(cat.flux/cat.dflux)
			freq_best   = np.argmax(freq_snr,1)
			best_contam = cat.contam[np.arange(len(cat)), freq_best]
			old_bad = bad.copy()
			bad  &= best_contam > 1/self.grid_dominance
			cat, cat_bad = cat[~bad], cat[bad]
			# Update model to reflect the cut sources
			if len(cat_bad) > 0:
				model -= self.modeller(cat_bad)
			print("cut %d srcs, %d remaining" % (len(cat_bad), len(cat)))
		return bunch.Bunch(cat=cat, snmin=snmin, model=model, hits=hits, bad_mask=bad_mask)

# Measurer implementations

class MeasurerSimple(Measurer):
	def __init__(self, nmat, beam, scaling=1):
		self.beam   = beam
		self.nmat   = nmat
		self.scaling= scaling
		self.order  = 3
	def __call__(self, map, icat):
		assert map.ndim == 3, "Map must be an enmap with 3 dims"
		ncomp = len(map)
		cat   = icat.copy()
		pixs  = map.sky2pix([icat.dec,icat.ra])
		# Apply the matched filter
		rho, kappa = self.nmat.matched_filter(map, self.beam)
		kappa     = sanitize_kappa(kappa)
		# Read off the total values at the given positions
		scaling   = np.zeros(len(rho),rho.dtype)+self.scaling
		rho_tot   = np.sum(rho*scaling[:,None,None],0)
		kappa_tot = np.sum(scaling[:,None,None,None]*kappa*scaling[None,:,None,None],(0,1))
		snr_tot   = rho_tot/kappa_tot**0.5
		flux_tot, dflux_tot = solve_mapsys(kappa_tot, rho_tot)
		del rho_tot, kappa_tot
		cat.snr             = snr_tot  .at(pixs, unit="pix", order=0)
		cat.flux_tot        = flux_tot .at(pixs, unit="pix", order=self.order)
		cat.dflux_tot       = dflux_tot.at(pixs, unit="pix", order=0)
		del snr_tot, flux_tot, dflux_tot
		# Read off the individual values
		flux, dflux = solve_mapsys(kappa, rho)
		del rho, kappa
		cat.flux    = flux .at(pixs, unit="pix", order=self.order).T
		cat.dflux   = dflux.at(pixs, unit="pix", order=0).T
		return bunch.Bunch(cat=cat)

class MeasurerMulti(Measurer):
	def __init__(self, measurers):
		self.measurers = measurers
	def __call__(self, map, icat):
		cat = icat.copy()
		if len(icat) == 0: return bunch.Bunch(cat=cat)
		uvals, order, edges = utils.find_equal_groups_fast(icat.case)
		for i, ca in enumerate(uvals):
			subicat = icat[order[edges[i]:edges[i+1]]]
			if len(subicat) == 0: continue
			subocat = self.measurers[i](map, subicat).cat
			cat[order[edges[i]:edges[i+1]]] = subocat
		return bunch.Bunch(cat=cat)

class MeasurerIterative(Measurer):
	def __init__(self, measurer, modeller, sntol=0.25, snscale=1):
		self.measurer = measurer
		self.modeller = modeller
		self.sntol    = sntol
		self.snscale  = snscale
		self.snmin    = 0.1 # do everything at once below this
	def __call__(self, map, icat, verbose=False):
		cat    = icat.copy()
		if cat.size == 0: return bunch.Bunch(cat=cat, model=self.modeller(cat))
		snr    = icat.snr * self.snscale
		groups = snr_split(snr, sntol=self.sntol, snmin=self.snmin)
		model  = np.zeros_like(map)
		for gi, group in enumerate(groups):
			if verbose: print("Measuring group %d with snmin %6.2f" % (gi+1, np.min(snr[group])))
			subcat = self.measurer(map-model, icat[group]).cat
			model += self.modeller(subcat)
			cat[group] = subcat
		return bunch.Bunch(cat=cat, model=model)


# Modeller implementations

class ModellerPerfreq(Modeller):
	def __init__(self, shape, wcs, beam_profiles, dtype=np.float32, nsigma=5):
		self.shape = shape
		self.wcs   = wcs
		self.dtype = dtype
		self.nsigma= nsigma
		self.beam_profiles = []
		for i, (r,b) in enumerate(beam_profiles):
			self.beam_profiles.append(np.array([r,b/np.max(b)]))
		self.areas = np.array([utils.calc_beam_area(prof) for prof in self.beam_profiles])
	def __call__(self, cat):
		from . import pointsrcs
		ncomp = len(self.beam_profiles)
		omap  = enmap.zeros((ncomp,)+self.shape[-2:], self.wcs, self.dtype)
		if len(cat) == 0: return omap
		for i in range(ncomp):
			# This just subtracts the raw measurement at each frequencies. Some frequencies may have
			# bad S/N. Maybe consider adding the scaled tot flux as a weak prior?
			srcparam = np.concatenate([cat.dec[:,None],cat.ra[:,None],cat.flux[:,i:i+1]/self.areas[i]],-1)
			pointsrcs.sim_srcs(self.shape[-2:], self.wcs, srcparam, self.beam_profiles[i], omap=omap[i], nsigma=self.nsigma)
		return omap
	def amplitudes(self, cat):
		bpeaks = np.array([prof[1,0] for prof in self.beam_profiles])
		return cat.flux * (bpeaks/self.areas)

class ModellerScaled(Modeller):
	def __init__(self, shape, wcs, beam_profiles, scaling, dtype=np.float32, nsigma=5):
		self.shape = shape
		self.wcs   = wcs
		self.dtype = dtype
		self.nsigma= nsigma
		self.scaling = scaling
		self.beam_profiles = []
		for i, (r,b) in enumerate(beam_profiles):
			self.beam_profiles.append(np.array([r,b/np.max(b)]))
		self.areas = np.array([utils.calc_beam_area(prof) for prof in self.beam_profiles])
	def __call__(self, cat):
		from . import pointsrcs
		ncomp = len(self.beam_profiles)
		omap  = enmap.zeros((ncomp,)+self.shape[-2:], self.wcs, self.dtype)
		if len(cat) == 0: return omap
		for i in range(ncomp):
			srcparam = np.concatenate([cat.dec[:,None],cat.ra[:,None],cat.flux_tot[:,None]*self.scaling[i]/self.areas[i]],-1)
			pointsrcs.sim_srcs(self.shape[-2:], self.wcs, srcparam, self.beam_profiles[i], omap=omap[i], nsigma=self.nsigma)
		return omap
	def amplitudes(self, cat):
		bpeaks = np.array([prof[1,0] for prof in self.beam_profiles])
		return cat.flux_tot[:,None]*(self.scaling*bpeaks/self.areas)

class ModellerMulti(Modeller):
	def __init__(self, modellers):
		self.modellers = modellers
	def __call__(self, cat):
		# If the cat is empty, just pass it on to the first modeller to have
		# it generate an empty map
		if len(cat) == 0: return self.modellers[0](cat)
		# Loop through the catalog entries of each type and have the corresponding
		# modeller build a model for them
		uvals, order, edges = utils.find_equal_groups_fast(cat.case)
		omap = None
		for i, ca in enumerate(uvals):
			subcat = cat[order[edges[i]:edges[i+1]]]
			if len(subcat) == 0: continue
			map    = self.modellers[ca](subcat)
			if omap is None: omap  = map
			else:            omap += map
		return omap
	def amplitudes(self, cat):
		res = np.zeros(cat.flux.shape)
		if len(cat) == 0: return res
		uvals, order, edges = utils.find_equal_groups_fast(cat.case)
		for i, ca in enumerate(uvals):
			subcat = cat[order[edges[i]:edges[i+1]]]
			res[order[edges[i]:edges[i+1]]] = self.modellers[ca].amplitudes(subcat)
		return res

# Helper functions. Should maybe be moved to utils

def sanitize_kappa(kappa, tol=1e-4, inplace=False):
	if not inplace: kappa = kappa.copy()
	for i in range(len(kappa)):
		kappa[i,i] = np.maximum(kappa[i,i], np.max(kappa[i,i])*tol)
	return kappa

def solve_mapsys(kappa, rho):
	if kappa.ndim == 2:
		flux  = rho/kappa
		dflux = kappa**-0.5
	else:
		# Check if this is slow
		flux  = enmap.samewcs(np.linalg.solve(kappa.T, rho.T).T, rho)
		dflux = enmap.samewcs(np.einsum("aayx->ayx",np.linalg.inv(kappa.T).T)**0.5, kappa)
	return flux, dflux

def get_flat_sky_correction(pixratio):
		return (0.5*(1+pixratio**2))**-0.5, 1/pixratio

def dtype_concat(dtypes):
	# numpy isn't cooperating, so I'm making this function myself
	return sum([np.dtype(dtype).descr for dtype in dtypes],[])

def merge_arrays(arrays):
	odtype = dtype_concat([a.dtype for a in arrays])
	res    = np.zeros(arrays[0].shape, odtype)
	for a in arrays:
		for key in a.dtype.names:
			res[key] = a[key]
	return res

def rpow(fmap, exp=2):
	"""Given a fourier-space map fmap corresponding to a real map, take it to the given exponent in
	*real space*, and return the fourier-space version of the result."""
	norm = fmap.area()**0.5
	map  = enmap.ifft(fmap/norm+0j, normalize="phys").real
	return enmap.fft (map**exp, normalize="phys").real*norm

def rmul(*args):
	norm = args[0].area()**0.5
	work = None
	for arg in args:
		rmap = enmap.ifft(arg/norm+0j, normalize="phys").real
		if work is None: work  = rmap
		else:            work *= rmap
	return enmap.fft(work, normalize="phys").real*norm

def rop(*args, op=np.multiply):
	norm = args[0].area()**0.5
	return enmap.fft(op(*[enmap.ifft(arg/norm+0j, normalize="phys").real for arg in args]), normalize="phys").real*norm

def snr_split(snrs, sntol=0.25, snmin=5):
	"""Given a list of S/N ratios, split split them into groups that
	can be processed together without interfering with each other.
	Returns [inds1, inds2, inds3, ...], where inds1 has the indices
	of the strongest snrs, inds2 slightly weaker snrs, and so on.
	The weakest element in a group will be at least sntol times the strongest
	element. Values below snmin are bunched into a single group."""
	v  = np.log(np.maximum(np.abs(snrs), snmin))/np.log(1/sntol)
	v -= np.max(v)+1e-9
	v  = utils.floor(v)
	return utils.find_equal_groups(v)[::-1]

def get_ref(a, tol=1e-3, default=0, n=1000):
	ref  = 0
	for i in range(2):
		vals = a[a>ref]
		if vals.size == 0: return default
		step = max(1,vals.size//n)
		ref  = np.median(vals[::step])
	return ref

def safe_pow(N, pow, bad_tol=1e-3):
	from enlib import array_ops
	v   = np.einsum("aa...->a...", N)
	ref = np.array([get_ref(vi) for vi in v])
	return array_ops.eigpow(N, pow, axes=[0,1], lim0=ref*bad_tol)

def make_circle_labels(shape, wcs, pixs, inds=None, r=2*utils.arcmin):
	if inds is None: inds = np.arange(1,len(pixs[0])+1)
	mask  = enmap.zeros(shape[-2:], wcs, np.int32)
	mask[pixs[0],pixs[1]] = inds
	dists, labels = mask.labeled_distance_transform(rmax=r)
	labels[dists >= r] = 0
	return labels

def get_central_radius(fbeam, lknee=2000, alpha=-3):
	"""Given a fourier-space beam and a filter lknee, get the radius at which
	the beam falls to zero for the first time. If would be more
	exact if this used the actual noise model, but in that case
	it's hard to know at what pixel to evaluate it, given that
	parts of the map could be empty."""
	l = fbeam.modlmap()
	fbeam = np.mean(fbeam.preflat,0)
	with utils.nowarn():
		fbeam = fbeam * (1 + (l/lknee)**alpha)**-1
	rbeam = enmap.ifft(fbeam+0j).real
	pos   = fbeam.pix2sky([0,0])
	br, r = rbeam.rbin(pos)
	br   /= br[0]
	return r[np.nonzero(br < 0)[0][0]]

#def fbeam_to_area(fbeam):
#	beam  = enmap.ifft(fbeam+0j).real
#	beam /= np.max(beam)
#	return np.sum(beam)*beam.pixsize()

# This huge comment was in NmatConstcorr. I've put it here for now
# in case it's interesting, but I'll probably delete it eventually.

# Our model that our map contains a single point source + noise,
#  m = BPa+n
# where a is the total flux in that pixel, P is something that puts all that flux in the correct
# pixel, and B is our beam, which preserves total flux but smears it out. So B[l=0] = 1. This means
# that P should have units of 1/pixarea so that Pa has units of flux per steradian.
#
# Our ML estimate of a is then
#  a = (P'B'N"BP)"P'B'N"m = rhs/kappa
#  rhs = P'B'N"m, kappa = P'B'N"BP
# In real space, each column of B would be a set of numbers that sum to 1. Near the equator
# these would be more concentrated, further away they would be broader. The peak height would
# be position-independent. The pixel size wouldn't enter here (aside from deciding how many
# pixels are hit) since the flux density is an intensive quantity.
#
# Regardless of projection B(x,y) would look like b(dist(x,y)), which is symmetric.
#
# How does the flat sky approximation enter here? Affects both B and N, but N doesn't affect
# the flux expectation value, so let's ignore it for now. Flat sky means that we replace B
# with B2(x,y) = b(dist0(x,y)), dist0(x,y) = ((x.ra-y.ra)**2*cos_dec + (x.dec-y.dec)**2)**0.5.
#
# What is (P'B2'B2P)"P'B2'BP? 
#
# BP  = a beam map centered on some location dec with total flux of 1. In car it's stretched hor by 1/cos(dec)
# B2P = similar, but stretched hor by 1/cos(dec0).
# (B2P)'(BP) is the dot product of these. Let's try for a gaussian:
# int 1/sqrt(2pi s1**2) 1/sqrt(2pi s2**2) exp(-(x*s1)**2) exp(-(x*s2)**2) dx =
# int 1/... exp(-(x*(s1**2+s2**2))) dx = sqrt(2 pi (s1**2+s2**2))/sqrt(2 pi s1**2)/sqrt(2 pi s2**2)
# = 1/sqrt(2 pi) * sqrt(s1**-2+s2**-2)
# What it should have been:
# 1/sqrt(2 pi) * sqrt(2 s1**-2)
#
# So (P'B2'B2P)"P'B2'BP = sqrt(s1**-2+s2**-2)/sqrt(s2**-2+s2**-2)
# where it should have been 1 if B2 were B. So we're off by a factor
# sqrt(cos(dec)**-2 + cos(dec0)**-2)/(2 cos(dec0)**-2)
#
# Hm, here I assumed that the normalization was different for the two cases,
# but is it? No, we have the same peak in all cases. The flat sky approximation
# only affects how we compute distances. So let's try again:
#
#   int exp(-0.5*(x*s1)**2)*exp(-0.5*(x*s2)**2) dx
# = int exp(-0.5*x**2*(s1**2+s2**2))
# = sqrt(2*pi*(s1**2+s2**2))
#
# (P'B2'B2P)"P'B2'P = sqrt(s1**2+s2**2)/sqrt(s2**2+s2**2)
# For my case that means we make a flux error of
# sqrt(cos(dec)**2+cos(dec0))/sqrt(2 cos(dec0)**2)
# = sqrt(0.5*(1+cos(dec)**2/cos(dec0)**2))
#
# This depends on it being gaussian, but let's just use it for now.
# What happens to rho and kappa separately?
# rho is wrong by sqrt(s1**2+s2**2)/sqrt(2*s1**2)
# = sqrt(cos(dec)**2+cos(dec0)**2)/sqrt(2*cos(dec)**2)
# = sqrt(0.5*(1+(cos(dec)/cos(dec0))**-2))
# kappa is wrong by sqrt(2 s2**2)/sqrt(2 s1**2) = (cos(dec)/cos(dec0))**-1
#
# But wait! We've forgotten about P! P should go as 1/pixsizemap, but
# in the flat sky it goes as pixsize. So we should keep P and P2 separate.
#
# Let's try again:
#
# rho error is:
#  rho2/rho = P2'B2'BP/(P'B'BP) = sqrt(0.5*(1+(cos(dec)/cos(dec0))**-2)) * area/area0
#           = sqrt(0.5*(1+(area/area0)**-2)) * (area/area0)
#           = sqrt(0.5*(1+(area/area0)**2))
# kappa error is
#  kappa2/kappa = (P2'B2'B2P2)'/(P'B'BP) = (cos(dec)/cos(dec0))**-1 * (area0/area)**-2
#           = area/area0



"""Utilities for making mpi use safer and easier."""
from __future__ import print_function
import sys, os, traceback, copy, numpy as np
from .mpiutils import FakeCommunicator, FAKE_WORLD

COMM_WORLD = FAKE_WORLD
COMM_SELF  = FAKE_WORLD
disabled   = True

# Allow us to disable real mpi, creating only a simple placeholder object
# that will work for trivially parallelizable code run with only 1 task
try:
    if not("DISABLE_MPI" in os.environ and os.environ["DISABLE_MPI"].lower() in ["true","1"]):
        # We transparently pass through all the mpi4py.MPI stuff, but also add a cleanup
        # hook. On scinet I found that uncaught exceptions did not cause mpi to abort,
        # leading to thousands of wasted CPU hours. That may have been system-specific,
        # though. Perhaps this isn't necessary in general.
        from mpi4py.MPI import *
        #def cleanup(type, value, traceback):
        #    sys.__excepthook__(type, value, traceback)
        #    COMM_WORLD.Abort(1)
        #sys.excepthook = cleanup
        disabled = False
except:
    pass
